from .lineage import (
    Lineage,
    SubclassesIterator,
    SubpropertiesIterator,
    SuperclassesIterator,
    SuperpropertiesIterator,
)

__all__ = [
    Lineage.__name__,
    SubclassesIterator.__name__,
    SubpropertiesIterator.__name__,
    SuperclassesIterator.__name__,
    SuperpropertiesIterator.__name__,
]



import datetime
import os
import re
import typing
import warnings
from typing import Dict, List, Optional, Tuple
from xml.etree import ElementTree as etree

import dateutil.parser

from ..definition import Definition
from ..metadata import Metadata, Subset
from ..pv import LiteralPropertyValue, ResourcePropertyValue
from ..synonym import SynonymData, SynonymType
from ..utils.meta import typechecked
from ..utils.warnings import NotImplementedWarning, SyntaxWarning
from ..xref import Xref
from .base import BaseParser

if typing.TYPE_CHECKING:
    from ..entity import Entity
    from ..ontology import Ontology


class Namespace(object):
    def __init__(self, base: str):
        self.base = base

    def __getitem__(self, item: str) -> str:
        return f"{{{self.base}}}{item}"

    def raw(self, item: str) -> str:
        return f"{self.base}{item}"


_NS = {
    "dc": Namespace("http://purl.org/dc/elements/1.1/"),
    "doap": Namespace("http://usefulinc.com/ns/doap#"),
    "foaf": Namespace("http://xmlns.com/foaf/0.1/"),
    "meta": Namespace("http://www.co-ode.org/ontologies/meta.owl#"),
    "obo": Namespace("http://purl.obolibrary.org/obo/"),
    "oboInOwl": Namespace("http://www.geneontology.org/formats/oboInOwl#"),
    "owl": Namespace("http://www.w3.org/2002/07/owl#"),
    "protege": Namespace("http://protege.stanford.edu/plugins/owl/protege#"),
    "rdf": Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#"),
    "rdfs": Namespace("http://www.w3.org/2000/01/rdf-schema#"),
    "skos": Namespace("http://www.w3.org/2004/02/skos/core#"),
    "ubprop": Namespace("http://purl.obolibrary.org/obo/ubprop#"),
    "uberon": Namespace("http://purl.obolibrary.org/obo/uberon#"),
    "xsd": Namespace("http://www.w3.org/2001/XMLSchema#"),
}

_SYNONYMS = {
    _NS["oboInOwl"].raw("hasExactSynonym"): "EXACT",
    _NS["oboInOwl"].raw("hasBroadSynonym"): "BROAD",
    _NS["oboInOwl"].raw("hasNarrowSynonym"): "NARROW",
    _NS["oboInOwl"].raw("hasRelatedSynonym"): "RELATED",
}

_SYNONYMS_ATTRIBUTES = {
    _NS["oboInOwl"]["hasExactSynonym"]: "EXACT",
    _NS["oboInOwl"]["hasBroadSynonym"]: "BROAD",
    _NS["oboInOwl"]["hasNarrowSynonym"]: "NARROW",
    _NS["oboInOwl"]["hasRelatedSynonym"]: "RELATED",
}


class RdfXMLParser(BaseParser):
    """A parser for OWL2 ontologies serialized in RDF/XML format.

    To Do:
        * Extraction of annotation properties, which are currently simply
          being ignored when parsing a document.
    """

    # -- BaseParser interface ------------------------------------------------

    @classmethod
    def can_parse(cls, path, buffer):
        return buffer.lstrip().startswith((b"<?xml", b"<rdf:RDF", b"<owl:"))

    def parse_from(self, handle, threads=None):
        # Load the XML document into an XML Element tree
        tree: etree.ElementTree = etree.parse(handle)

        # Load metadata from the `owl:Ontology` element
        owl_ontology = tree.find(_NS["owl"]["Ontology"])
        if owl_ontology is None:
            raise ValueError("could not find `owl:Ontology` element")
        with typechecked.disabled():
            self.ont.metadata = self._extract_meta(owl_ontology)

        # Process imports
        self.ont.imports.update(
            self.process_imports(
                self.ont.metadata.imports,
                self.ont.import_depth,
                os.path.dirname(self.ont.path or str()),
                self.ont.timeout,
                threads=threads,
            )
        )

        # Merge lineage cache from imports
        self.import_lineage()

        # Disable typecheck to make parsing faster
        with typechecked.disabled():

            curies = self._make_curies(tree)

            for prop in tree.iterfind(_NS["owl"]["ObjectProperty"]):
                self._extract_object_property(prop, curies)
            for prop in tree.iterfind(_NS["owl"]["AnnotationProperty"]):
                self._extract_annotation_property(prop, curies)
            for class_ in tree.iterfind(_NS["owl"]["Class"]):
                self._extract_term(class_, curies)
            for axiom in tree.iterfind(_NS["owl"]["Axiom"]):
                self._process_axiom(axiom, curies)

        # Update lineage cache with symmetric of `subClassOf`
        self.symmetrize_lineage()

    # -- Helper methods ------------------------------------------------------

    def _make_curies(self, tree: etree.ElementTree):
        curies = {}
        # Extract `oboInOwl:id` or compact IRI for `owl:Class`
        for elem in tree.iterfind(_NS["owl"]["Class"]):
            iri = elem.get(_NS["rdf"]["about"])
            if iri is None:
                continue
            e = elem.find(_NS["oboInOwl"]["id"])
            id_: str = e.text if e is not None and e.text else self._compact_id(iri)
            curies[iri] = id_
        # Extract either `oboInOwl:shorthand` or `oboInOwl:id` for `owl:ObjectProperty`
        for elem in tree.iterfind(_NS["owl"]["ObjectProperty"]):
            iri = elem.get(_NS["rdf"]["about"])
            if iri is None:
                continue
            # attempt to extract the compact id of the term
            elem_id = elem.find(_NS["oboInOwl"]["id"])
            elem_sh = elem.find(_NS["oboInOwl"]["shorthand"])
            if elem_sh is not None and elem_sh.text is not None:
                curies[iri] = elem_sh.text
            elif elem_id is not None and elem_id.text is not None:
                curies[iri] = elem_id.text
            else:
                curies[iri] = self._compact_id(iri)
        # Compact IRI of owl:AnnotationProperty
        for elem in tree.iterfind(_NS["owl"]["AnnotationProperty"]):
            iri = elem.get(_NS["rdf"]["about"])
            if iri is None:
                continue
            # attempt to extract the compact id of the term
            elem_id = elem.find(_NS["oboInOwl"]["id"])
            elem_sh = elem.find(_NS["oboInOwl"]["shorthand"])
            if elem_sh is not None and elem_sh.text is not None:
                curies[iri] = elem_sh.text
            elif elem_id is not None and elem_id.text is not None:
                curies[iri] = elem_id.text
            else:
                curies[iri] = self._compact_id(iri)
        return curies

    def _compact_id(self, iri: str) -> str:
        """Compact an OBO identifier into a prefixed identifier."""
        match = re.match("^http://purl.obolibrary.org/obo/([^#_]+)_(.*)$", iri)
        if match is not None:
            return ":".join(match.groups())
        if self.ont.metadata.ontology is not None:
            id_ = self.ont.metadata.ontology
            match = re.match(f"^http://purl.obolibrary.org/obo/{id_}#(.*)$", iri)
            if match is not None:
                return match.group(1)
        return iri

    def _compact_datatype(self, iri: str) -> str:
        match = re.match("^http://www.w3.org/2001/XMLSchema#(.*)$", iri)
        if match is not None:
            return f"xsd:{match.group(1)}"
        raise ValueError(f"invalid datatype: {iri!r}")

    def _extract_term_relationship(
        self, elem: etree.Element
    ) -> Tuple[Optional[str], Optional[str]]:
        """Extract the relationship info from a `owl:Restriction` element.

        This only handles the simplest case of `owl:someValuesFrom` annotation
        and not extra cases described in the `class expression paragraph of the
        OBO guide <http://owlcollab.github.io/oboformat/doc/obo-syntax.html#5.3>`_.
        """
        if __debug__:
            if elem.tag != _NS["owl"]["Restriction"]:
                raise ValueError("expected `owl:Restriction` element")

        prop = elem.find(_NS["owl"]["onProperty"])
        iri_prop = prop.get(_NS["rdf"]["resource"]) if prop is not None else None
        if iri_prop is None:
            warnings.warn(
                "could not extract property IRI from `owl:Restriction`",
                SyntaxWarning,
                stacklevel=3,
            )
            return None, None

        target = elem.find(_NS["owl"]["someValuesFrom"])
        iri_target = target.get(_NS["rdf"]["resource"]) if target is not None else None
        if iri_target is None:
            warnings.warn(
                "could not extract target IRI from `owl:Restriction`",
                SyntaxWarning,
                stacklevel=3,
            )
            return None, None

        # TODO: check for `ObjectExactCardinality`, `ObjectAllValuesFrom`, etc.
        return iri_prop, iri_target

    def _extract_resource_pv(self, elem: etree.Element) -> ResourcePropertyValue:
        property = re.sub("{|}", "", elem.tag)
        resource = elem.attrib[_NS["rdf"]["resource"]]
        return ResourcePropertyValue(property, resource)

    def _extract_literal_pv(self, elem: etree.Element) -> LiteralPropertyValue:
        property = re.sub("{|}", "", elem.tag)
        datatype = elem.get(_NS["rdf"]["datatype"])
        if datatype is None:
            warnings.warn(
                f"{elem} contains text but no `xsd:datatype`",
                SyntaxWarning,
                stacklevel=2,
            )
            datatype = _NS["xsd"].raw("string")
        return LiteralPropertyValue(
            property, typing.cast(str, elem.text), self._compact_datatype(datatype)
        )

    def _extract_meta(self, elem: etree.Element):
        """Extract the metadata from an `owl:Ontology` element."""
        meta = Metadata()
        if __debug__:
            if elem.tag != _NS["owl"]["Ontology"]:
                raise ValueError("expected `owl:Ontology` element")

        # extract OBO format version
        iri = elem.get(_NS["rdf"]["about"])
        if iri is not None:
            match = re.match("^http://purl.obolibrary.org/obo/(.*).(obo|owl)$", iri)
            meta.ontology = match.group(1) if match is not None else iri

        # extract metadata from child elements
        for child in elem:
            if child.tag == _NS["rdfs"]["comment"] and child.text is not None:
                meta.remarks.add(child.text)
            elif child.tag == _NS["oboInOwl"]["hasOBOFormatVersion"]:
                meta.format_version = child.text
            elif child.tag in (_NS["oboInOwl"]["saved-by"], _NS["oboInOwl"]["savedBy"]):
                meta.saved_by = child.text
            elif child.tag == _NS["oboInOwl"]["auto-generated-by"]:
                meta.auto_generated_by = child.text
            elif child.tag in (
                _NS["oboInOwl"]["default-namespace"],
                _NS["oboInOwl"]["hasDefaultNamespace"],
            ):
                meta.default_namespace = child.text
            elif child.tag == _NS["oboInOwl"]["date"]:
                meta.date = datetime.datetime.strptime(child.text, "%d:%m:%Y %H:%M")
            elif child.tag == _NS["oboInOwl"]["NamespaceIdRule"]:
                meta.namespace_id_rule = child.text
            elif child.tag == _NS["owl"]["imports"]:
                meta.imports.add(child.get(_NS["rdf"]["resource"]))
            elif child.tag == _NS["owl"]["versionIRI"]:
                iri = child.get(_NS["rdf"]["resource"])
                if iri is not None and meta.ontology is not None:
                    rx = "^http://purl.obolibrary.org/obo/{0}/(.*)/{0}.(obo|owl)$"
                    match = re.match(rx.format(meta.ontology), iri)
                else:
                    match = None
                meta.data_version = iri if match is None else match.group(1)
            elif child.tag == _NS["doap"]["Version"]:
                meta.data_version = child.text
            elif _NS["rdf"]["resource"] in child.attrib:
                meta.annotations.add(self._extract_resource_pv(child))
            elif child.text is not None:
                meta.annotations.add(self._extract_literal_pv(child))
            else:
                warnings.warn(
                    f"unknown element in `owl:Ontology`: {child}",
                    SyntaxWarning,
                    stacklevel=3,
                )

        # return the extracted metadata
        return meta

    def _extract_term(self, elem: etree.Element, curies: Dict[str, str]):
        """Extract the term from a `owl:Class` element."""
        if __debug__:
            if elem.tag != _NS["owl"]["Class"]:
                raise ValueError("expected `owl:Class` element")

        # only create the term if it is not a class by restriction
        iri: Optional[str] = elem.get(_NS["rdf"]["about"])
        if iri is None:
            return None

        # get the CURIE corresponding to the term
        id_ = curies[iri]

        # get or create the term
        term = (self.ont.get_term if id_ in self.ont.terms() else self.ont.create_term)(id_)
        termdata = term._data()
        names: List[str] = []
        comments: List[str] = []

        # extract attributes from annotation of the OWL class
        for child in elem:

            tag: str = child.tag
            attrib: Dict[str, str] = child.attrib
            text: Optional[str] = child.text
            if text is not None and text.isspace():
                text = None

            if tag == _NS["rdfs"]["subClassOf"]:
                restriction = child.find(_NS["owl"]["Restriction"])
                if _NS["rdf"]["resource"] in attrib:
                    if attrib[_NS["rdf"]["resource"]] != _NS["owl"].raw("Thing"):
                        iri = attrib[_NS["rdf"]["resource"]]
                        curie = curies.get(iri) or self._compact_id(iri)
                        self.ont._terms.lineage[id_].sup.add(curie)
                elif restriction is not None:
                    r, t = self._extract_term_relationship(restriction)
                    if r is not None and t is not None:
                        r, t = curies.get(r, r), self._compact_id(t)
                        termdata.relationships.setdefault(r, set()).add(t)
                else:
                    warnings.warn(
                        "cannot process `rdfs:subClassOf` in this context",
                        SyntaxWarning,
                        stacklevel=2,
                    )
            elif tag == _NS["oboInOwl"]["inSubset"]:
                iri = attrib.get(_NS["rdf"]["resource"], text)
                if iri is not None:
                    curie = curies.get(iri) or self._compact_id(iri)
                    termdata.subsets.add(curie)
                else:
                    warnings.warn(
                        f"could not extract subset value in {id_!r}",
                        SyntaxWarning,
                        stacklevel=2,
                    )
            elif tag == _NS["rdfs"]["comment"] and text is not None:
                comments.append(text)
            elif tag in (_NS["oboInOwl"]["created_by"], _NS["dc"]["creator"]):
                termdata.created_by = text
            elif tag in (_NS["oboInOwl"]["creation_date"], _NS["dc"]["date"]):
                termdata.creation_date = dateutil.parser.parse(typing.cast(str, text))
            elif tag == _NS["oboInOwl"]["hasOBONamespace"]:
                if text != self.ont.metadata.default_namespace:
                    termdata.namespace = text
            elif tag == _NS["rdfs"]["label"]:
                if text is not None:
                    names.append(text)
                else:
                    warnings.warn(
                        f"`rdfs:label` without text literal in {id!r}",
                        SyntaxWarning,
                        stacklevel=2,
                    )
            elif tag == _NS["obo"]["IAO_0000115"] and text is not None:
                termdata.definition = Definition(text)
            elif tag in _SYNONYMS_ATTRIBUTES:
                scope = _SYNONYMS_ATTRIBUTES[tag]
                description = attrib.get(_NS["rdf"]["resource"], text)
                if description is not None:
                    termdata.synonyms.add(SynonymData(description, scope))
                else:
                    warnings.warn(
                        f"could not extract synonym value in {id_!r}",
                        SyntaxWarning,
                        stacklevel=3,
                    )
            elif tag == _NS["owl"]["equivalentClass"] and text is not None:
                curie = curies.get(text) or self._compact_id(text)
                termdata.equivalent_to.add(curie)
            elif tag == _NS["owl"]["deprecated"]:
                termdata.obsolete = text == "true"
            elif tag == _NS["oboInOwl"]["hasDbXref"]:
                try:
                    if text is not None:
                        termdata.xrefs.add(Xref(text))
                    else:
                        termdata.xrefs.add(Xref(attrib[_NS["rdf"]["resource"]]))
                except ValueError:
                    pass
            elif tag == _NS["oboInOwl"]["hasAlternativeId"]:
                iri = attrib.get(_NS["rdf"]["resource"], text)
                curie = curies.get(iri) or self._compact_id(iri)
                termdata.alternate_ids.add(curie)
            elif tag == _NS["owl"]["disjointWith"]:
                if _NS["rdf"]["resource"] in attrib:
                    iri = attrib[_NS["rdf"]["resource"]]
                    curie = curies.get(iri) or self._compact_id(iri)
                    termdata.disjoint_from.add(curie)
                else:
                    warnings.warn(
                        "`owl:disjointWith` element without `rdf:resource`",
                        SyntaxWarning,
                        stacklevel=2,
                    )
            elif tag == _NS["obo"]["IAO_0100001"]:
                if _NS["rdf"]["resource"] in attrib:
                    iri = attrib[_NS["rdf"]["resource"]]
                    curie = curies.get(iri) or self._compact_id(iri)
                    termdata.replaced_by.add(curie)
                elif _NS["rdf"]["datatype"] in attrib:
                    curie = curies.get(text) or self._compact_id(text)
                    termdata.replaced_by.add(curie)
                else:
                    warnings.warn(
                        "could not extract ID from `IAO:0100001` annotation",
                        SyntaxWarning,
                        stacklevel=2,
                    )
            elif tag == _NS["oboInOwl"]["consider"]:
                if _NS["rdf"]["resource"] in attrib:
                    iri = attrib[_NS["rdf"]["resource"]]
                    curie = curies.get(iri) or self._compact_id(iri)
                    termdata.consider.add(curie)
                elif _NS["rdf"]["datatype"] in attrib:
                    curie = curies.get(text) or self._compact_id(text)
                    termdata.consider.add(curie)
                else:
                    warnings.warn(
                        "could not extract ID from `oboInOwl:consider` annotation",
                        SyntaxWarning,
                        stacklevel=3,
                    )
            elif tag != _NS["oboInOwl"]["id"]:
                if _NS["rdf"]["resource"] in attrib:
                    termdata.annotations.add(self._extract_resource_pv(child))
                elif _NS["rdf"]["datatype"] in attrib and text is not None:
                    termdata.annotations.add(self._extract_literal_pv(child))
                else:
                    warnings.warn(
                        f"unknown element in `owl:Class`: {child.tag}",
                        SyntaxWarning,
                        stacklevel=2,
                    )

        # Owl to OBO post processing:
        # see http://owlcollab.github.io/oboformat/doc/obo-syntax.html#5.11
        # check we got a single name, or select an arbitrary one
        if names:
            if len(names) > 1:
                warnings.warn(
                    f"several names found for {id_!r}, using {names[0]!r}",
                    SyntaxWarning,
                    stacklevel=2,
                )
            termdata.name = names[0]
        # check we got a single comment, or concatenate comments
        if comments:
            if len(comments) > 1:
                warnings.warn(
                    f"several names found for {id_!r}, concatenating",
                    SyntaxWarning,
                    stacklevel=2,
                )
            termdata.comment = "\n".join(comments)

    def _extract_object_property(self, elem: etree.Element, curies: Dict[str, str]):
        """Extract the object property from an `owl:ObjectProperty` element."""
        if __debug__:
            if elem.tag != _NS["owl"]["ObjectProperty"]:
                raise ValueError("expected `owl:ObjectProperty` element")

        # only create the term if it is not a restriction
        iri: Optional[str] = elem.get(_NS["rdf"]["about"])
        if iri is None:  # ignore
            return None

        # get the CURIE of the typedef
        id_ = curies[iri]

        # create the relationship
        rel = (
            self.ont.get_relationship
            if id_ in self.ont.relationships()
            else self.ont.create_relationship
        )(id_)
        reldata = rel._data()
        names: List[str] = []
        comments: List[str] = []

        # extract attributes from annotation of the OWL relationship
        for child in elem:

            tag: str = child.tag
            attrib: Dict[str, str] = child.attrib
            text: Optional[str] = child.text
            if text is not None and text.isspace():
                text = None

            if tag == _NS["rdfs"]["subPropertyOf"]:
                if _NS["rdf"]["resource"] in attrib:
                    iri = attrib[_NS["rdf"]["resource"]]
                    if iri != _NS["owl"].raw("topObjectProperty"):
                        curie = curies.get(iri) or self._compact_id(iri)
                        self.ont._relationships.lineage[id_].sup.add(curie)
                else:
                    pass  # TODO: subclassing relationship for relationship
            elif tag == _NS["oboInOwl"]["inSubset"]:
                resource = child.get(_NS["rdf"]["resource"])
                about = child.get(_NS["rdf"]["about"])
                if resource or about:
                    reldata.subsets.add(self._compact_id(resource or about))
                else:
                    warnings.warn(
                        f"could not extract subset in {id_!r}",
                        SyntaxWarning,
                        stacklevel=3,
                    )
            elif tag == _NS["rdf"]["type"]:
                resource = child.get(_NS["rdf"]["resource"])
                if resource == _NS["owl"].raw("TransitiveProperty"):
                    reldata.transitive = True
                elif resource == _NS["owl"].raw("ReflexiveProperty"):
                    reldata.reflexive = True
                elif resource == _NS["owl"].raw("SymmetricProperty"):
                    reldata.symmetric = True
                elif resource == _NS["owl"].raw("AsymmetricProperty"):
                    reldata.asymmetric = True
                elif resource == _NS["owl"].raw("FunctionalProperty"):
                    reldata.functional = True
                elif resource == _NS["owl"].raw("InverseFunctionalProperty"):
                    reldata.inverse_functional = True
            elif tag == _NS["rdfs"]["comment"] and text is not None:
                comments.append(text)
            elif tag in (_NS["oboInOwl"]["created_by"], _NS["dc"]["creator"]):
                reldata.created_by = text
            elif (
                tag == _NS["oboInOwl"]["creation_date"] or tag == _NS["dc"]["date"]
            ) and text is not None:
                reldata.creation_date = dateutil.parser.parse(text)
            elif tag == _NS["oboInOwl"]["hasOBONamespace"]:
                if text != self.ont.metadata.default_namespace:
                    reldata.namespace = text
            elif tag == _NS["rdfs"]["label"]:
                if text is not None:
                    names.append(text)
                else:
                    warnings.warn(
                        f"`rdfs:label` without text literal in {id!r}",
                        SyntaxWarning,
                        stacklevel=2,
                    )
            elif tag == _NS["rdfs"]["domain"] and _NS["rdf"]["resource"] in attrib:
                reldata.domain = self._compact_id(child.attrib[_NS["rdf"]["resource"]])
            elif tag == _NS["rdfs"]["range"] and _NS["rdf"]["resource"] in attrib:
                reldata.range = self._compact_id(child.attrib[_NS["rdf"]["resource"]])
            elif tag == _NS["obo"]["IAO_0000115"] and text is not None:
                reldata.definition = Definition(text)
            elif tag in _SYNONYMS_ATTRIBUTES:
                scope = _SYNONYMS_ATTRIBUTES[tag]
                description = child.get(_NS["rdf"]["resource"], text)
                reldata.synonyms.add(SynonymData(description, scope))
            elif tag == _NS["oboInOwl"]["is_cyclic"] and text is not None:
                reldata.cyclic = text == "true"
            elif tag == _NS["obo"]["IAO_0000427"] and text is not None:
                reldata.antisymmetric = text == "true"
            elif tag == _NS["owl"]["equivalentClass"] and text is not None:
                curie = curies.get(text) or self._compact_id(text)
                reldata.equivalent_to.add(curie)
            elif tag == _NS["owl"]["inverseOf"]:
                iri = child.attrib[_NS["rdf"]["resource"]]
                reldata.inverse_of = curies.get(iri) or self._compact_id(iri)
            elif tag == _NS["owl"]["deprecated"]:
                reldata.obsolete = text == "true"
            elif tag == _NS["oboInOwl"]["hasDbXref"]:
                if text is not None:
                    reldata.xrefs.add(Xref(text))
                else:
                    reldata.xrefs.add(Xref(attrib[_NS["rdf"]["resource"]]))
            elif tag == _NS["oboInOwl"]["hasAlternativeId"] and text is not None:
                reldata.alternate_ids.add(text)
            elif tag == _NS["obo"]["IAO_0100001"]:
                if _NS["rdf"]["resource"] in attrib:
                    iri = attrib[_NS["rdf"]["resource"]]
                    curie = curies.get(iri) or self._compact_id(iri)
                    reldata.replaced_by.add(curie)
                elif _NS["rdf"]["datatype"] in attrib:
                    curie = curies.get(text) or self._compact_id(text)
                    reldata.replaced_by.add(curie)
                else:
                    warnings.warn(
                        "could not extract ID from IAO:0100001 annotation",
                        SyntaxWarning,
                        stacklevel=3,
                    )
            elif tag == _NS["oboInOwl"]["consider"]:
                if _NS["rdf"]["resource"] in attrib:
                    iri = attrib[_NS["rdf"]["resource"]]
                    curie = curies.get(iri) or self._compact_id(iri)
                    reldata.consider.add(curie)
                elif _NS["rdf"]["datatype"] in attrib:
                    curie = curies.get(text) or self._compact_id(text)
                    reldata.consider.add(curie)
                else:
                    warnings.warn(
                        "could not extract ID from `oboInOwl:consider` annotation",
                        SyntaxWarning,
                        stacklevel=2,
                    )
            elif tag not in (_NS["oboInOwl"]["id"], _NS["oboInOwl"]["shorthand"]):
                if _NS["rdf"]["resource"] in attrib:
                    reldata.annotations.add(self._extract_resource_pv(child))
                elif _NS["rdf"]["datatype"] in attrib and text is not None:
                    reldata.annotations.add(self._extract_literal_pv(child))
                else:
                    warnings.warn(
                        f"unknown element in `owl:ObjectProperty`: {child}",
                        SyntaxWarning,
                        stacklevel=2,
                    )

        # Owl to OBO post processing:
        # see http://owlcollab.github.io/oboformat/doc/obo-syntax.html#5.11
        # check we got a single name, or select an arbitrary one
        if names:
            if len(names) > 1:
                warnings.warn(
                    f"several names found for {id_!r}, using {names[0]!r}",
                    SyntaxWarning,
                    stacklevel=2,
                )
            reldata.name = names[0]
        # check we got a single comment, or concatenate comments
        if comments:
            if len(comments) > 1:
                warnings.warn(
                    f"several names found for {id_!r}, concatenating",
                    SyntaxWarning,
                    stacklevel=3,
                )
            reldata.comment = "\n".join(comments)

        return rel

    def _extract_annotation_property(self, elem: etree.Element, curies: Dict[str, str]):
        if __debug__:
            if elem.tag != _NS["owl"]["AnnotationProperty"]:
                raise ValueError("expected `owl:AnnotationProperty` element")

        #
        iri = elem.attrib[_NS["rdf"]["about"]]
        id_ = curies[iri]

        # special handling of `synonymtypedef` and `subsetdef`
        sub = elem.find(_NS["rdfs"]["subPropertyOf"])
        if sub is not None:
            resource = sub.get(_NS["rdf"]["resource"])
            if resource == _NS["oboInOwl"].raw("SynonymTypeProperty"):
                # extract ID and label of the synonymtypedef
                label = elem.find(_NS["rdfs"]["label"]).text
                if label is None:
                    label = ""
                    warnings.warn(
                        "`oboInOwl:SynonymTypeProperty` element has no label",
                        SyntaxWarning,
                        stacklevel=3,
                    )
                # extract scope if any
                elem_scope = elem.find(_NS["oboInOwl"]["hasScope"])
                if elem_scope is not None:
                    scope = _SYNONYMS.get(elem_scope.attrib[_NS["rdf"]["resource"]])
                else:
                    scope = None
                # add the synonymtypedef to the metadata
                self.ont.metadata.synonymtypedefs.add(SynonymType(id_, label, scope))
                return
            elif resource == _NS["oboInOwl"].raw("SubsetProperty"):
                elem_comment = elem.find(_NS["rdfs"]["comment"])
                desc = elem_comment.text if elem_comment is not None else None
                self.ont.metadata.subsetdefs.add(Subset(id_, desc or ""))
                return

        # TODO: actual annotation properties
        warnings.warn(
            "cannot process plain `owl:AnnotationProperty`",
            NotImplementedWarning,
            stacklevel=3,
        )

    def _process_axiom(self, elem: etree.Element, curies: Dict[str, str]):
        # get the source, property and target of the axiom.
        elem_source = elem.find(_NS["owl"]["annotatedSource"])
        elem_property = elem.find(_NS["owl"]["annotatedProperty"])
        elem_target = elem.find(_NS["owl"]["annotatedTarget"])

        # assert source, target and property have a `rdf:resource` attribute.
        for x in (elem_source, elem_property):
            if x is None or _NS["rdf"]["resource"] not in x.attrib:
                return

        # check among known properties
        property = elem_property.attrib[_NS["rdf"]["resource"]]
        if property == _NS["obo"].raw("IAO_0000115") and elem_target.text is not None:
            iri = elem_source.attrib[_NS["rdf"]["resource"]]
            curie = curies.get(iri) or self._compact_id(iri)

            if curie in self.ont.terms():
                entity: Entity = self.ont.get_term(curie)
            elif curie in self.ont.relationships():
                entity: Entity = self.ont.get_relationship(curie)
            else:
                warnings.warn(
                    f"could not find axiom source: {curie!r}",
                    SyntaxWarning,
                    stacklevel=3,
                )
                return

            entity.definition = d = Definition(elem_target.text)
            for child in elem.iterfind(_NS["oboInOwl"]["hasDbXref"]):
                if child.text is not None:
                    try:
                        d.xrefs.add(Xref(child.text))
                    except ValueError:
                        warnings.warn(
                            f"could not parse Xref: {child.text!r}",
                            SyntaxWarning,
                            stacklevel=3,
                        )
                elif _NS["rdf"]["resource"] in child.attrib:
                    d.xrefs.add(Xref(child.get(_NS["rdf"]["resource"])))
                else:
                    warnings.warn(
                        "`oboInOwl:hasDbXref` element has no text",
                        SyntaxWarning,
                        stacklevel=3,
                    )

        elif (
            property == _NS["oboInOwl"].raw("hasDbXref")
            and elem_target.text is not None
        ):
            iri = elem_source.attrib[_NS["rdf"]["resource"]]
            entity = self.ont[curies.get(iri) or self._compact_id(iri)]
            label = elem.find(_NS["rdfs"]["label"])

            try:
                if label is not None and label.text is not None:
                    new_xref = Xref(elem_target.text, label.text)
                else:
                    new_xref = Xref(elem_target.text)
            except ValueError:
                warnings.warn(
                    f"could not parse Xref: {elem_target.text!r}",
                    SyntaxWarning,
                    stacklevel=3,
                )
                return

            if new_xref in entity._data().xrefs:
                entity._data().xrefs.remove(new_xref)
            entity._data().xrefs.add(new_xref)

        elif property in _SYNONYMS:
            iri = elem_source.attrib[_NS["rdf"]["resource"]]
            curie = curies.get(iri) or self._compact_id(iri)

            if curie in self.ont.terms():
                entity: Entity = self.ont.get_term(curie)
            elif curie in self.ont.relationships():
                entity: Entity = self.ont.get_relationship(curie)
            else:
                warnings.warn(
                    f"could not find axiom source: {curie!r}",
                    SyntaxWarning,
                    stacklevel=3,
                )
                return

            type_ = elem.find(_NS["oboInOwl"]["hasSynonymType"])
            try:
                synonym = next(
                    s._data()
                    for s in entity.synonyms
                    if s.description == elem_target.text
                    and s.scope == _SYNONYMS[property]
                )
            except StopIteration:
                description = elem_target.get(_NS["rdf"]["resource"], elem_target.text)
                if description is None:
                    warnings.warn(
                        f"could not extract synonym value in {elem!r}",
                        SyntaxWarning,
                        stacklevel=3,
                    )
                    return
                synonym = SynonymData(
                    description,
                    scope=_SYNONYMS[property],
                    type=type_.text if type_ is not None else None,
                )

            entity._data().synonyms.add(typing.cast(SynonymData, synonym))
            for child in elem.iterfind(_NS["oboInOwl"]["hasDbXref"]):
                if child.text is not None:
                    try:
                        synonym.xrefs.add(Xref(child.text))
                    except ValueError:
                        warnings.warn(
                            f"could not parse Xref: {child.text!r}",
                            SyntaxWarning,
                            stacklevel=3,
                        )
                else:
                    warnings.warn(
                        "`oboInOwl:hasDbXref` element has no text",
                        SyntaxWarning,
                        stacklevel=3,
                    )

        else:

            warnings.warn(
                f"unknown axiom property: {property!r}",
                SyntaxWarning,
                stacklevel=3,
            )



import typing
from typing import Iterable, Optional

from .utils.meta import roundrepr
from .xref import Xref

__all__ = ["Definition"]


class Definition(str):
    """A human-readable text definition of an entity.

    Definitions are human-readable descriptions of an entity in the ontology
    graph, with some optional cross-references to support the definition.

    Example:
        Simply create a `Definition` instance by giving it a string::

        >>> def1 = pronto.Definition('a structural anomaly')

        Additional cross-references can be passed as arguments, or added later
        to the ``xrefs`` attribute of the `Definition`:

        >>> def2 = pronto.Definition('...', xrefs={pronto.Xref('MGI:Anna')})
        >>> def2.xrefs.add(pronto.Xref('ORCID:0000-0002-3947-4444'))

        The text content of the definition can be accessed by casting the
        definition object to a plain string:

        >>> str(def1)
        'a structural anomaly'

    Caution:
        A `Definition` compare only based on its textual value, independently
        of the `Xref` it may contains:

        >>> def2 == pronto.Definition('...')
        True

    Note:
        Some ontologies use the xrefs of a description to attribute the
        authorship of that definition:

        >>> cio = pronto.Ontology.from_obo_library("cio.obo")
        >>> sorted(cio['CIO:0000011'].definition.xrefs)
        [Xref('Bgee:fbb')]

        The common usecase however is to refer to the source of a definition
        using persistent identifiers like ISBN book numbers or PubMed IDs.

        >>> pl = pronto.Ontology.from_obo_library("plana.obo")
        >>> sorted(pl['PLANA:0007518'].definition.xrefs)
        [Xref('ISBN:0-71677033-4'), Xref('PMID:4853064')]

    """

    xrefs: typing.Set[Xref]

    __slots__ = ("__weakref__", "xrefs")

    def __new__(cls, text: str, xrefs=None) -> "Definition":
        return super().__new__(cls, text)  # type: ignore

    def __init__(self, text: str, xrefs: Optional[Iterable[Xref]] = None) -> None:
        self.xrefs = set(xrefs) if xrefs is not None else set()

    def __repr__(self) -> str:
        return roundrepr.make("Definition", str(self), xrefs=(self.xrefs, set()))



"""A Python frontend to ontologies.

**pronto** is a Python agnostic library designed to work with ontologies. At
the moment, it can parse ontologies in the OBO, OBO Graphs or OWL in RDF/XML
format, on either the local host or from an network location, and export
ontologies to OBO or OBO Graphs (in JSON format).

Caution:
    Only classes and modules reachable from the top-level package ``pronto``
    are considered public and are guaranteed stable over `Semantic Versioning
    <https://semver.org/>`_. Use submodules (other than `~pronto.warnings`)
    at your own risk!

Note:
    ``pronto`` implements proper *type checking* for most of the methods and
    properties exposed in the public classes. This reproduces the behaviour
    of the Python standard library, to avoid common errors. This feature does
    however increase overhead, but can be disabled by executing Python in
    optimized mode (with the ``-O`` flag). *Parsing performances are not
    affected.*

"""

from .definition import Definition
from .entity import Entity
from .metadata import Metadata, Subset
from .ontology import Ontology
from .pv import LiteralPropertyValue, PropertyValue, ResourcePropertyValue
from .relationship import Relationship, RelationshipData, RelationshipSet
from .synonym import Synonym, SynonymData, SynonymType
from .term import Term, TermData, TermSet
from .utils import warnings
from .xref import Xref

# Using `__name__` attribute instead of directly using the name as a string
# so the linter doesn't complaint about unused imports in the top module
__all__ = [
    # modules
    "warnings",
    # classes
    Ontology.__name__,
    Entity.__name__,
    Term.__name__,
    TermData.__name__,
    TermSet.__name__,
    Metadata.__name__,
    Subset.__name__,
    Definition.__name__,
    Relationship.__name__,
    RelationshipData.__name__,
    RelationshipSet.__name__,
    Synonym.__name__,
    SynonymData.__name__,
    SynonymType.__name__,
    PropertyValue.__name__,
    LiteralPropertyValue.__name__,
    ResourcePropertyValue.__name__,
    Xref.__name__,
]

__author__ = "Martin Larralde <martin.larralde@embl.de>"
__license__ = "MIT"
__version__ = "2.4.6"

# Update the docstring with a link to the right version of the documentation
# instead of the latest.
if __doc__ is not None:
    __doc__ += f"""See Also:
    Online documentation for this version of the library on
    `Read The Docs <https://pronto.readthedocs.io/en/v{__version__}/>`_
    """



"""Cross-reference object definition.
"""

import typing

import fastobo

from .utils.meta import roundrepr, typechecked


__all__ = ["Xref"]


@roundrepr
class Xref(object):
    """A cross-reference to another document or resource.

    Cross-references (xrefs for short) can be used to back-up definitions of
    entities, synonyms, or to link ontological entities to other resources
    they may have been derived from. Although originally intended to provide
    links to databases, cross-references in OBO ontologies gained additional
    purposes, such as helping for header macros expansion, or being used to
    alias external relationships with local unprefixed IDs.

    The OBO format version 1.4 expects references to be proper OBO identifiers
    that can be translated to actual IRIs, which is a breaking change from the
    previous format. Therefore, cross-references are encouraged to be given as
    plain IRIs or as prefixed IDs using an ID from the IDspace mapping defined
    in the header.

    Example:
        A cross-reference in the Mammalian Phenotype ontology linking a term
        to some related Web resource:

        >>> mp = pronto.Ontology.from_obo_library("mp.obo")
        >>> mp["MP:0030151"].name
        'abnormal buccinator muscle morphology'
        >>> mp["MP:0030151"].xrefs
        frozenset({Xref('https://en.wikipedia.org/wiki/Buccinator_muscle')})

    Caution:
        `Xref` instances compare only using their identifiers; this means it
        is not possible to have several cross-references with the same
        identifier and different descriptions in the same set.

    Todo:
        Make sure to resolve header macros for xrefs expansion (such as
        ``treat-xrefs-as-is_a``) when creating an ontology, or provide a
        method on `~pronto.Ontology` doing so when called.

    """

    id: str
    description: typing.Optional[str]

    __slots__ = ("__weakref__", "id", "description")  # noqa: E0602

    @typechecked()
    def __init__(self, id: str, description: typing.Optional[str] = None):
        """Create a new cross-reference.

        Arguments:
            id (str): the identifier of the cross-reference, either as a URL,
                a prefixed identifier, or an unprefixed identifier.
            description (str or None): a human-readable description of the
                cross-reference, if any.

        """
        # check the id is valid using fastobo
        if not fastobo.id.is_valid(id):
            raise ValueError("invalid identifier: {}".format(id))

        self.id: str = id
        self.description = description

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Xref):
            return self.id == other.id
        return False

    def __gt__(self, other: object) -> bool:
        if isinstance(other, Xref):
            return self.id > other.id
        return NotImplemented

    def __ge__(self, other: object) -> bool:
        if isinstance(other, Xref):
            return self.id >= other.id
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Xref):
            return self.id < other.id
        return NotImplemented

    def __le__(self, other: object) -> bool:
        if isinstance(other, Xref):
            return self.id <= other.id
        return NotImplemented

    def __hash__(self):
        return hash(self.id)



import bz2
import codecs
import gzip
import io
import lzma
import typing
import urllib.request
import warnings
from http.client import HTTPResponse
from typing import BinaryIO, ByteString, Dict, Optional, Union, cast

import chardet

MAGIC_GZIP = bytearray([0x1F, 0x8B])
MAGIC_LZMA = bytearray([0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00, 0x00])
MAGIC_BZIP2 = bytearray([0x42, 0x5A, 0x68])


class BufferedReader(io.BufferedReader):
    """A patch for `io.BufferedReader` supporting `http.client.HTTPResponse`."""

    def read(self, size: Optional[int] = -1) -> bytes:
        try:
            return super(BufferedReader, self).read(size)
        except ValueError:
            if typing.cast(io.BufferedReader, self.closed):
                return b""
            raise


class EncodedFile(codecs.StreamRecoder):
    def __init__(
        self,
        file: BinaryIO,
        data_encoding: str,
        file_encoding: Optional[str] = None,
        errors: str = "strict",
    ):
        if file_encoding is None:
            file_encoding = data_encoding
        data_info = codecs.lookup(data_encoding)
        file_info = codecs.lookup(file_encoding)
        super().__init__(
            file,
            data_info.encode,
            data_info.decode,
            file_info.streamreader,
            file_info.streamwriter,
            errors,
        )
        # Add attributes to simplify introspection
        self.data_encoding = data_encoding
        self.file_encoding = file_encoding

    def read(self, size: Optional[int] = -1) -> bytes:
        chunk = super().read(-1 if size is None else size)
        return chunk.replace(b"\r\n", b"\n")

    def readinto(self, buffer: ByteString) -> int:
        chunk = self.read(len(buffer) // 2)
        typing.cast(bytearray, buffer)[: len(chunk)] = chunk
        return len(chunk)


def get_handle(path: str, timeout: int = 2) -> BinaryIO:
    """Given a path or URL, get a binary handle for that path."""
    try:
        return open(path, "rb", buffering=0)
    except Exception as err:
        headers = {"Keep-Alive": f"timeout={timeout}"}
        request = urllib.request.Request(path, headers=headers)
        res: HTTPResponse = urllib.request.urlopen(request, timeout=timeout)
        if not res.status == 200:
            raise ValueError(f"could not open {path}: {res.status} ({res.msg})")
        if res.headers.get("Content-Encoding") in {"gzip", "deflate"}:
            f = gzip.GzipFile(filename=res.geturl(), mode="rb", fileobj=res)
            return typing.cast(BinaryIO, f)
        return res


def get_location(reader: BinaryIO) -> Optional[str]:
    """Given a binary file-handle, try to extract the path/URL to the file."""
    return (
        getattr(reader, "name", None)
        or getattr(reader, "url", None)
        or getattr(reader, "geturl", lambda: None)()
    )


def decompress(
    reader: io.RawIOBase, path: Optional[str] = None, encoding: Optional[str] = None
) -> BinaryIO:
    """Given a binary file-handle, decompress it if it is compressed."""
    buffered = BufferedReader(reader)

    # Decompress the stream if it is compressed
    if buffered.peek().startswith(MAGIC_GZIP):
        decompressed = BufferedReader(
            typing.cast(
                io.RawIOBase,
                gzip.GzipFile(mode="rb", fileobj=typing.cast(BinaryIO, buffered)),
            )
        )
    elif buffered.peek().startswith(MAGIC_LZMA):
        decompressed = BufferedReader(
            typing.cast(
                io.RawIOBase, lzma.LZMAFile(typing.cast(BinaryIO, buffered), mode="rb")
            )
        )
    elif buffered.peek().startswith(MAGIC_BZIP2):
        decompressed = BufferedReader(
            typing.cast(
                io.RawIOBase, bz2.BZ2File(typing.cast(BinaryIO, buffered), mode="rb")
            )
        )
    else:
        decompressed = buffered

    # Attempt to detect the encoding and decode the stream
    det: Dict[str, Union[str, float]] = chardet.detect(decompressed.peek())
    confidence = 1.0 if encoding is not None else cast(float, det["confidence"])
    encoding = encoding if encoding is not None else cast(str, det["encoding"])

    if encoding == "ascii":
        encoding = "utf-8"
    if confidence < 1.0:
        warnings.warn(
            f"unsound encoding, assuming {encoding} ({confidence:.0%} confidence)",
            UnicodeWarning,
            stacklevel=3,
        )

    if encoding == "utf-8":
        return typing.cast(BinaryIO, decompressed)
    else:
        return typing.cast(
            BinaryIO,
            BufferedReader(
                typing.cast(
                    io.RawIOBase,
                    EncodedFile(
                        typing.cast(typing.BinaryIO, decompressed),
                        "UTF-8",
                        typing.cast(str, det["encoding"]),
                    ),
                )
            ),
        )



import abc
import functools
import multiprocessing.pool
import operator
import os
import typing
import urllib.parse
from typing import Dict, Optional, Set

from ..logic.lineage import Lineage
from ..ontology import Ontology


class BaseParser(abc.ABC):
    def __init__(self, ont: Ontology):
        self.ont: Ontology = ont

    @classmethod
    @abc.abstractmethod
    def can_parse(cls, path: str, buffer: bytes) -> bool:
        """Return `True` if this parser type can parse the given handle."""
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def parse_from(
        self, handle: typing.BinaryIO, threads: Optional[int] = None
    ) -> None:
        return NotImplemented

    @classmethod
    def process_import(
        cls,
        ref: str,
        import_depth: int = -1,
        basepath: str = "",
        timeout: int = 5,
    ) -> Ontology:
        s = urllib.parse.urlparse(ref).scheme
        if s in {"ftp", "http", "https"} or os.path.exists(ref):
            url = ref
        else:
            for ext in ["", ".obo", ".json", ".owl"]:
                if os.path.exists(os.path.join(basepath, f"{ref}{ext}")):
                    url = os.path.join(basepath, f"{ref}{ext}")
                    break
            else:
                id_ = f"{ref}.obo" if not os.path.splitext(ref)[1] else ref
                url = f"http://purl.obolibrary.org/obo/{id_}"
        return Ontology(url, max(import_depth - 1, -1), timeout)

    @classmethod
    def process_imports(
        cls,
        imports: Set[str],
        import_depth: int = -1,
        basepath: str = "",
        timeout: int = 5,
        threads: Optional[int] = None,
    ) -> Dict[str, Ontology]:
        # check we did not reach the maximum import depth
        if import_depth == 0:
            return dict()
        process = functools.partial(
            cls.process_import,
            import_depth=import_depth,
            basepath=basepath,
            timeout=timeout,
        )
        with multiprocessing.pool.ThreadPool(threads) as pool:
            return dict(pool.map(lambda i: (i, process(i)), imports))

    _entities = {
        "Term": operator.attrgetter("terms", "_terms"),
        "Relationship": operator.attrgetter("relationships", "_relationships"),
    }

    def symmetrize_lineage(self):
        for getter in self._entities.values():
            entities, graphdata = getter(self.ont)
            for entity in entities():
                graphdata.lineage.setdefault(entity.id, Lineage())
            for subentity, lineage in graphdata.lineage.items():
                for superentity in lineage.sup:
                    graphdata.lineage[superentity].sub.add(subentity)

    def import_lineage(self):
        for getter in self._entities.values():
            entities, graphdata = getter(self.ont)
            for dep in self.ont.imports.values():
                dep_entities, dep_graphdata = getter(dep)
                for entity in dep_entities():
                    graphdata.lineage[entity.id] = Lineage()
                for id, lineage in dep_graphdata.lineage.items():
                    graphdata.lineage[id].sup.update(lineage.sup)
                    graphdata.lineage[id].sub.update(lineage.sub)



from .base import BaseSerializer
from .obo import OboSerializer
from .obojson import OboJSONSerializer
from .ofn import OwlFunctionalSerializer

__all__ = [
    "BaseSerializer",
    "OboSerializer",
    "OboJSONSerializer",
    "OwlFunctionalSerializer"
]



import operator
import typing

import fastobo

from ..metadata import Metadata
from ..ontology import Ontology
from ..pv import LiteralPropertyValue, PropertyValue, ResourcePropertyValue
from ..relationship import Relationship, RelationshipData
from ..synonym import SynonymData
from ..term import Term, TermData
from ..xref import Xref


class FastoboSerializer:

    ont: Ontology

    def _to_obodoc(self, o: Ontology) -> fastobo.doc.OboDoc:
        doc = fastobo.doc.OboDoc()
        if o.metadata:
            doc.header = self._to_header_frame(o.metadata)
        for termdata in sorted(
            self.ont._terms.entities.values(), key=operator.attrgetter("id")
        ):
            doc.append(self._to_term_frame(Term(self.ont, termdata)))
        for reldata in sorted(
            self.ont._relationships.entities.values(), key=operator.attrgetter("id")
        ):
            doc.append(self._to_typedef_frame(Relationship(self.ont, reldata)))
        return doc

    def _to_header_frame(self, m: Metadata) -> fastobo.header.HeaderFrame:
        # Ordering of tags follow the OBO 1.4 specification
        frame = fastobo.header.HeaderFrame()
        if m.format_version is not None:
            frame.append(fastobo.header.FormatVersionClause(m.format_version))
        if m.data_version is not None:
            frame.append(fastobo.header.DataVersionClause(m.data_version))
        if m.date is not None:
            frame.append(fastobo.header.DateClause(m.date))
        if m.saved_by is not None:
            frame.append(fastobo.header.SavedByClause(m.saved_by))
        if m.auto_generated_by is not None:
            frame.append(fastobo.header.AutoGeneratedByClause(m.auto_generated_by))
        for i in sorted(m.imports):
            frame.append(fastobo.header.ImportClause(i))
        for subset in sorted(m.subsetdefs):
            frame.append(
                fastobo.header.SubsetdefClause(
                    subset=fastobo.id.parse(subset.name), description=subset.description
                )
            )
        for syn in sorted(m.synonymtypedefs):
            frame.append(
                fastobo.header.SynonymTypedefClause(
                    typedef=fastobo.id.parse(syn.id),
                    description=syn.description,
                    scope=syn.scope,
                )
            )
        if m.default_namespace is not None:
            frame.append(fastobo.header.DefaultNamespaceClause(m.default_namespace))
        if m.namespace_id_rule is not None:
            frame.append(fastobo.header.NamespaceIdRuleClause(m.namespace_id_rule))
        for id, (url, description) in sorted(m.idspaces.items()):
            frame.append(fastobo.header.IdspaceClause(id, fastobo.id.Url(url), description))
        for pv in sorted(m.annotations):
            frame.append(
                fastobo.header.PropertyValueClause(self._to_property_value(pv))
            )
        for remark in sorted(m.remarks):
            frame.append(fastobo.header.RemarkClause(remark))
        if m.ontology is not None:
            frame.append(fastobo.header.OntologyClause(m.ontology))
        for line in m.owl_axioms:
            frame.append(fastobo.header.OwlAxiomsClause(line))
        for tag, values in sorted(m.unreserved.items()):
            for value in values:
                frame.append(fastobo.header.UnreservedClause(tag, value))
        return frame

    def _to_property_value(self, pv: PropertyValue) -> fastobo.pv.AbstractPropertyValue:
        try:
            pv = typing.cast(ResourcePropertyValue, pv)
            return fastobo.pv.ResourcePropertyValue(
                fastobo.id.parse(pv.property),
                fastobo.id.parse(pv.resource),
            )
        except AttributeError:
            pv = typing.cast(LiteralPropertyValue, pv)
            return fastobo.pv.LiteralPropertyValue(
                fastobo.id.parse(pv.property), pv.literal, fastobo.id.parse(pv.datatype)
            )

    def _to_synonym(self, syn: SynonymData) -> fastobo.syn.Synonym:
        return fastobo.syn.Synonym(
            syn.description,
            syn.scope,
            None if syn.type is None else fastobo.id.parse(syn.type),
            map(self._to_xref, syn.xrefs),
        )

    def _to_term_frame(self, term: Term) -> fastobo.term.TermFrame:
        t = term._data()
        frame = fastobo.term.TermFrame(fastobo.id.parse(t.id))
        if t.anonymous:
            frame.append(fastobo.term.IsAnonymousClause(True))
        if t.name is not None:
            frame.append(fastobo.term.NameClause(t.name))
        if t.namespace is not None:
            if t.namespace != self.ont.metadata.default_namespace:
                ns = fastobo.id.parse(t.namespace)
                frame.append(fastobo.term.NamespaceClause(ns))
        for alt in sorted(t.alternate_ids):
            frame.append(fastobo.term.AltIdClause(fastobo.id.parse(alt)))
        if t.definition is not None:
            frame.append(
                fastobo.term.DefClause(
                    str(t.definition),
                    [self._to_xref(x) for x in sorted(t.definition.xrefs)],
                )
            )
        if t.comment is not None:
            frame.append(fastobo.term.CommentClause(t.comment))
        for subset in sorted(t.subsets):
            frame.append(fastobo.term.SubsetClause(fastobo.id.parse(subset)))
        for syn in sorted(t.synonyms):
            frame.append(fastobo.term.SynonymClause(self._to_synonym(syn)))
        for xref in sorted(t.xrefs):
            frame.append(fastobo.term.XrefClause(self._to_xref(xref)))
        if t.builtin:
            frame.append(fastobo.term.BuiltinClause(True))
        for pv in sorted(t.annotations):
            frame.append(fastobo.term.PropertyValueClause(self._to_property_value(pv)))
        for superclass in sorted(term.superclasses(with_self=False, distance=1)):
            frame.append(fastobo.term.IsAClause(fastobo.id.parse(superclass.id)))
        for i in sorted(filter(lambda x: not isinstance(x, tuple), t.intersection_of)):
            frame.append(fastobo.term.IntersectionOfClause(term=fastobo.id.parse(i), typedef=None))
        for (i, j) in sorted(filter(lambda x: isinstance(x, tuple), t.intersection_of)):  # type: ignore
            frame.append(
                fastobo.term.IntersectionOfClause(
                    typedef=fastobo.id.parse(i), term=fastobo.id.parse(j)
                )
            )
        for id_ in sorted(t.union_of):
            frame.append(fastobo.term.UnionOfClause(fastobo.id.parse(id_)))
        for id_ in sorted(t.equivalent_to):
            frame.append(fastobo.term.EquivalentToClause(fastobo.id.parse(id_)))
        for id_ in sorted(t.disjoint_from):
            frame.append(fastobo.term.DisjointFromClause(fastobo.id.parse(id_)))
        for r, values in t.relationships.items():
            r_id = fastobo.id.parse(r)
            for value in values:
                t_id = fastobo.id.parse(value)
                frame.append(fastobo.term.RelationshipClause(r_id, t_id))
        if t.created_by is not None:
            frame.append(fastobo.term.CreatedByClause(t.created_by))
        if t.creation_date is not None:
            frame.append(fastobo.term.CreationDateClause(t.creation_date))
        if t.obsolete:
            frame.append(fastobo.term.IsObsoleteClause(True))
        for r in sorted(t.replaced_by):
            frame.append(fastobo.term.ReplacedByClause(fastobo.id.parse(r)))
        for c in sorted(t.consider):
            frame.append(fastobo.term.ConsiderClause(fastobo.id.parse(c)))
        return frame

    def _to_typedef_frame(self, relationship: Relationship):
        r = relationship._data()
        frame = fastobo.typedef.TypedefFrame(fastobo.id.parse(r.id))
        if r.anonymous:
            frame.append(fastobo.typedef.IsAnonymousClause(True))
        if r.name is not None:
            frame.append(fastobo.typedef.NameClause(r.name))
        if r.namespace is not None:
            if r.namespace != self.ont.metadata.default_namespace:
                ns = fastobo.id.parse(r.namespace)
                frame.append(fastobo.typedef.NamespaceClause(ns))
        for alt in sorted(r.alternate_ids):
            frame.append(fastobo.typedef.AltIdClause(fastobo.id.parse(alt)))
        if r.definition is not None:
            frame.append(
                fastobo.typedef.DefClause(
                    str(r.definition),
                    [self._to_xref(x) for x in sorted(r.definition.xrefs)],
                )
            )
        if r.comment is not None:
            frame.append(fastobo.typedef.CommentClause(r.comment))
        for subset in sorted(r.subsets):
            frame.append(fastobo.typedef.SubsetClause(fastobo.id.parse(subset)))
        for syn in sorted(r.synonyms):
            frame.append(fastobo.typedef.SynonymClause(self._to_synonym(syn)))
        for xref in sorted(r.xrefs):
            frame.append(fastobo.typedef.XrefClause(self._to_xref(xref)))
        for pv in sorted(r.annotations):
            frame.append(
                fastobo.typedef.PropertyValueClause(self._to_property_value(pv))
            )
        if r.domain is not None:
            frame.append(fastobo.typedef.DomainClause(fastobo.id.parse(r.domain)))
        if r.range is not None:
            frame.append(fastobo.typedef.RangeClause(fastobo.id.parse(r.range)))
        if r.builtin:
            frame.append(fastobo.typedef.BuiltinClause(True))
        for chain in sorted(r.holds_over_chain):
            c1, c2 = map(fastobo.id.parse, chain)
            frame.append(fastobo.typedef.HoldsOverChainClause(c1, c2))
        if r.antisymmetric:
            frame.append(fastobo.typedef.IsAntiSymmetricClause(True))
        if r.cyclic:
            frame.append(fastobo.typedef.IsCyclicClause(True))
        if r.reflexive:
            frame.append(fastobo.typedef.IsReflexiveClause(True))
        if r.asymmetric:
            frame.append(fastobo.typedef.IsAsymmetricClause(True))
        if r.symmetric:
            frame.append(fastobo.typedef.IsSymmetricClause(True))
        if r.transitive:
            frame.append(fastobo.typedef.IsTransitiveClause(True))
        if r.functional:
            frame.append(fastobo.typedef.IsFunctionalClause(True))
        if r.inverse_functional:
            frame.append(fastobo.typedef.IsInverseFunctionalClause(True))
        for superproperty in sorted(
            relationship.superproperties(with_self=False, distance=1)
        ):
            frame.append(fastobo.typedef.IsAClause(fastobo.id.parse(superproperty.id)))
        for i in sorted(r.intersection_of):
            frame.append(fastobo.typedef.IntersectionOfClause(fastobo.id.parse(i)))
        for id_ in sorted(r.union_of):
            frame.append(fastobo.typedef.UnionOfClause(fastobo.id.parse(id_)))
        for id_ in sorted(r.equivalent_to):
            frame.append(fastobo.typedef.EquivalentToClause(fastobo.id.parse(id_)))
        for id_ in sorted(r.disjoint_from):
            frame.append(fastobo.typedef.DisjointFromClause(fastobo.id.parse(id_)))
        if r.inverse_of is not None:
            frame.append(
                fastobo.typedef.InverseOfClause(fastobo.id.parse(r.inverse_of))
            )
        for id_ in sorted(r.transitive_over):
            frame.append(fastobo.typedef.TransitiveOverClause(fastobo.id.parse(id_)))
        for chain in sorted(r.equivalent_to_chain):
            c1, c2 = map(fastobo.id.parse, chain)
            frame.append(fastobo.typedef.EquivalentToChainClause(c1, c2))
        for id_ in sorted(r.disjoint_over):
            frame.append(fastobo.typedef.DisjointOverClause(fastobo.id.parse(id_)))
        for rel, values in r.relationships.items():
            if rel != "is_a":
                r_id = fastobo.id.parse(rel)
                for value in values:
                    t_id = fastobo.id.parse(value)
                    frame.append(fastobo.typedef.RelationshipClause(r_id, t_id))
        if r.obsolete:
            frame.append(fastobo.typedef.IsObsoleteClause(True))
        if r.created_by is not None:
            frame.append(fastobo.typedef.CreatedByClause(r.created_by))
        if r.creation_date is not None:
            frame.append(fastobo.typedef.CreationDateClause(r.creation_date))
        for id_ in sorted(r.replaced_by):
            frame.append(fastobo.typedef.ReplacedByClause(fastobo.id.parse(id_)))
        for id_ in sorted(r.consider):
            frame.append(fastobo.typedef.ConsiderClause(fastobo.id.parse(id_)))
        for d in r.expand_assertion_to:
            frame.append(
                fastobo.typedef.ExpandAssertionToClause(
                    str(d),
                    [self._to_xref(x) for x in sorted(d.xrefs)],
                )
            )
        for d in r.expand_expression_to:
            frame.append(
                fastobo.typedef.ExpandExpressionToClause(
                    str(d),
                    [self._to_xref(x) for x in sorted(d.xrefs)],
                )
            )
        if r.metadata_tag:
            frame.append(fastobo.typedef.IsMetadataTagClause(True))
        if r.class_level:
            frame.append(fastobo.typedef.IsClassLevel(True))
        return frame

    def _to_xref(self, x: Xref) -> fastobo.xref.Xref:
        return fastobo.xref.Xref(fastobo.id.parse(x.id), x.description)



import datetime
import operator
import typing
import weakref
from typing import AbstractSet, Any, Dict, FrozenSet, Iterable, Iterator, Optional, Set

from ..definition import Definition
from ..pv import PropertyValue
from ..synonym import Synonym, SynonymData, SynonymType
from ..utils.meta import roundrepr, typechecked
from ..xref import Xref

if typing.TYPE_CHECKING:
    from ..ontology import _DataGraph, Ontology
    from ..relationship import Relationship, RelationshipSet
    from .attributes import Relationships

__all__ = ["EntityData", "Entity", "EntitySet"]
_D = typing.TypeVar("_D", bound="EntityData")
_E = typing.TypeVar("_E", bound="Entity")
_S = typing.TypeVar("_S", bound="EntitySet")


class EntityData:

    id: str
    alternate_ids: Set[str]
    annotations: Set[PropertyValue]
    anonymous: bool
    builtin: bool
    comment: Optional[str]
    consider: Set[str]
    created_by: Optional[str]
    creation_date: Optional[datetime.datetime]
    disjoint_from: Set[str]
    definition: Optional[Definition]
    equivalent_to: Set[str]
    name: Optional[str]
    namespace: Optional[str]
    obsolete: bool
    relationships: Dict[str, Set[str]]
    replaced_by: Set[str]
    subsets: Set[str]
    synonyms: Set[SynonymData]
    union_of: Set[str]
    xrefs: Set[Xref]

    if typing.TYPE_CHECKING:
        __annotations__: Dict[str, str]

    __slots__ = ("__weakref__",) + tuple(__annotations__)  # noqa: E0602


class Entity(typing.Generic[_D, _S]):
    """An entity in the ontology graph.

    With respects to the OBO semantics, an `Entity` is either a term or a
    relationship in the ontology graph. Any entity has a unique identifier as
    well as some common properties.
    """

    if __debug__ or typing.TYPE_CHECKING:

        __data: "weakref.ReferenceType[_D]"
        __slots__: Iterable[str] = ()

        def __init__(self, ontology: "Ontology", data: "_D"):
            self.__data = weakref.ref(data)
            self.__id = data.id
            self.__ontology = ontology

        def _data(self) -> "EntityData":
            rdata = self.__data()
            if rdata is None:
                raise RuntimeError("internal data was deallocated")
            return rdata

    else:

        __slots__: Iterable[str] = ("_data",)  # type: ignore

        def __init__(self, ontology: "Ontology", data: "_D"):
            self._data = weakref.ref(data)  # type: ignore
            self.__ontology = ontology
            self.__id = data.id

    _Set: typing.ClassVar[typing.Type[_S]] = NotImplemented
    _data_getter: typing.Callable[["Ontology"], "_DataGraph"] = NotImplemented

    # --- Private helpers ----------------------------------------------------

    def _ontology(self) -> "Ontology":
        return self.__ontology

    # --- Magic Methods ------------------------------------------------------

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Entity):
            return self.id == other.id
        return False

    def __lt__(self, other):
        if isinstance(other, Entity):
            return self.id < other.id
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, Entity):
            return self.id <= other.id
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Entity):
            return self.id > other.id
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Entity):
            return self.id >= other.id
        return NotImplemented

    def __hash__(self):
        return hash((self.id))

    def __repr__(self):
        return roundrepr.make(type(self).__name__, self.id, name=(self.name, None))

    # --- Data descriptors ---------------------------------------------------

    @property
    def alternate_ids(self) -> Set[str]:
        """`set` of `str`: A set of alternate IDs for this entity."""
        return self._data().alternate_ids

    @alternate_ids.setter  # type: ignore
    def alternate_ids(self, ids: Iterable[str]):
        self._data().alternate_ids = set(ids)

    @property
    def annotations(self) -> Set[PropertyValue]:
        """`set` of `PropertyValue`: Annotations relevant to the entity."""
        return self._data().annotations

    @annotations.setter
    def annotations(self, value: Iterable[PropertyValue]) -> None:
        self._data().annotations = set(value)

    @property
    def anonymous(self) -> bool:
        """`bool`: Whether or not the entity has an anonymous id.

        Semantics of anonymous entities are the same as B-Nodes in RDF.
        """
        return self._data().anonymous

    @anonymous.setter
    def anonymous(self, value: bool):
        self._data().anonymous = value

    @property
    def builtin(self) -> bool:
        """`bool`: Whether or not the entity is built-in to the OBO format.

        ``pronto`` uses this tag on the ``is_a`` relationship, which is the
        axiomatic to the OBO language but treated as a relationship in the
        library.
        """
        return self._data().builtin

    @builtin.setter  # type: ignore
    @typechecked(property=True)
    def builtin(self, value: bool):
        self._data().builtin = value

    @property
    def comment(self) -> Optional[str]:
        """`str` or `None`: A comment about the current entity.

        Comments in ``comment`` clauses are guaranteed to be conserved by OBO
        parsers and serializers, unlike bang comments. A non `None` `comment`
        is semantically equivalent to a ``rdfs:comment`` in OWL2. When parsing
        from OWL, several RDF comments will be merged together into a single
        ``comment`` clause spanning over multiple lines.
        """
        return self._data().comment

    @comment.setter
    def comment(self, value: Optional[str]):
        self._data().comment = value

    @property
    def consider(self) -> _S:
        """`EntitySet`: A set of potential substitutes for an obsolete term.

        An obsolete entity can provide one or more entities which may be
        appropriate substitutes, but needs to be looked at carefully by a
        human expert before the replacement is done.

        See Also:
            `~Entity.replaced_by`, which provides a set of entities suitable
            for automatic replacement.

        """
        s = self._Set()
        s._ids = self._data().consider
        s._ontology = self._ontology()
        return s

    @consider.setter
    def consider(self, consider: Iterable[_E]) -> None:
        if isinstance(consider, EntitySet):
            data = consider._ids
        else:
            data = {entity.id for entity in consider}
        self._data().consider = data

    @property
    def created_by(self) -> Optional[str]:
        """`str` or `None`: The name of the creator of the entity, if any.

        This property gets translated to a ``dc:creator`` annotation in OWL2,
        which has very broad semantics. Some OBO ontologies may instead use
        other annotation properties such as the ones found in `Information
        Interchange Ontology <http://www.obofoundry.org/ontology/iao.html>`_,
        which can be accessed in the `annotations` attribute of the entity,
        if any.
        """
        return self._data().created_by

    @created_by.setter  # type: ignore
    @typechecked(property=True)
    def created_by(self, value: Optional[str]):
        self._data().created_by = value

    @property
    def creation_date(self) -> Optional[datetime.datetime]:
        """`~datetime.datetime` or `None`: The date the entity was created."""
        return self._data().creation_date

    @creation_date.setter  # type: ignore
    @typechecked(property=True)
    def creation_date(self, value: Optional[datetime.datetime]):
        self._data().creation_date = value

    @property
    def definition(self) -> Optional[Definition]:
        """`Definition` or `None`: The definition of the current entity.

        Definitions in OBO are intended to be human-readable text describing
        the entity, with some additional cross-references if possible.

        Example:
            >>> hp = pronto.Ontology.from_obo_library("hp.obo")
            >>> term = hp["HP:0009882"]
            >>> term.name
            'Short distal phalanx of finger'
            >>> str(term.definition)
            'Short distance from the end of the finger to the most distal...'
            >>> sorted(term.definition.xrefs)
            [Xref('HPO:probinson'), Xref('PMID:19125433')]

        """
        return self._data().definition

    @definition.setter  # type: ignore
    @typechecked(property=True)
    def definition(self, definition: Optional[Definition]):
        self._data().definition = definition

    @property
    def disjoint_from(self) -> _S:
        """`EntitySet`: The entities declared as disjoint from this entity.

        Two entities are disjoint if they have no instances in common. Two
        entities that are disjoint cannot share any subentities, but the
        opposite is not always true.
        """
        s = self._Set()
        s._ids = self._data().disjoint_from
        s._ontology = self._ontology()
        return s

    @disjoint_from.setter
    def disjoint_from(self, disjoint: Iterable[_E]):
        if isinstance(disjoint, EntitySet):
            data = disjoint._ids
        else:
            data = {entity.id for entity in disjoint}
        self._data().disjoint_from = data

    @property
    def equivalent_to(self) -> _S:
        """`EntitySet`: The entities declared as equivalent to this entity."""
        s = self._Set()
        s._ids = self._data().equivalent_to
        s._ontology = self._ontology()
        return s

    @equivalent_to.setter
    def equivalent_to(self, entities: Iterable[_E]):
        if isinstance(entities, EntitySet):
            data = entities._ids
        else:
            data = {entity.id for entity in entities}
        self._data().equivalent_to = data

    @property
    def id(self) -> str:
        """`str`: The OBO identifier of the entity.

        Identifiers can be either prefixed (e.g. ``MS:1000031``), unprefixed
        (e.g. ``part_of``) or given as plain URLs. Identifiers cannot be
        edited.
        """
        return self.__id

    @property
    def name(self) -> Optional[str]:
        """`str` or `None`: The name of the entity.

        Names are formally equivalent to ``rdf:label`` in OWL2. The OBO format
        version 1.4 made names optional to improve OWL interoperability, as
        labels are optional in OWL.
        """
        return self._data().name

    @name.setter  # type: ignore
    @typechecked(property=True)
    def name(self, value: Optional[str]):
        self._data().name = value

    @property
    def namespace(self) -> Optional[str]:
        """`str` or `None`: The namespace this entity is defined in."""
        return self._data().namespace

    @namespace.setter  # type: ignore
    @typechecked(property=True)
    def namespace(self, ns: Optional[str]):
        self._data().namespace = ns

    @property
    def obsolete(self) -> bool:
        """`bool`: Whether or not the entity is obsolete.

        Hint:
            All OBO entities can be made obsolete through a boolean flag, and
            map to one or several replacements. When querying an obsolete
            entity, ``pronto`` will **not** attempt to perform any kind of
            replacement itself ::

                >>> ms = pronto.Ontology.from_obo_library("ms.obo")
                >>> term = ms["MS:1001414"]
                >>> term
                Term('MS:1001414', name='MGF scans')
                >>> term.obsolete
                True

            To always get the up-to-date, non-obsolete entity, you could use
            the following snippet, going through a term replacement if there
            is no ambiguity ::

                >>> while term.obsolete:
                ...     if len(term.replaced_by) != 1:
                ...         raise ValueError(f"no replacement for {term.id}")
                ...     term = term.replaced_by.pop()
                >>> term
                Term('MS:1000797', name='peak list scans')

        See Also:
            `~.Entity.consider` and `~Entity.replaced_by`, storing some
            replacement options for an obsolete entity.

        """
        return self._data().obsolete

    @obsolete.setter  # type: ignore
    @typechecked(property=True)
    def obsolete(self, value: bool):
        self._data().obsolete = value

    @property
    def relationships(self: _E) -> "Relationships[_E, _S]":
        """`~.Relationships`: The links from an entity to other entities.

        This property returns an object that maps a `~.Relationship` to
        an `~.EntitySet` (either a `~.TermSet` for `Term.relationships`, or
        a `~.RelationshipSet` for `Relationship.relationships`).

        Hint:
            The mapping is mutable, so relationships can be created or removed
            using the usual interface of a `~collections.abc.MutableMapping`.

        Example:
            Get the ``MS:1000004`` term (*sample mass*) from the Mass
            Spectrometry ontology::

                >>> ms = pronto.Ontology.from_obo_library("ms.obo")
                >>> sample_mass = ms["MS:1000004"]

            Then use the ``relationships`` property to get the relevant
            unit from the Unit Ontology::

                >>> sorted(sample_mass.relationships.keys())
                [Relationship('has_units', name='has_units')]
                >>> sample_mass.relationships[ms.get_relationship('has_units')]
                TermSet({Term('UO:0000021', name='gram')})

        """
        from .attributes import Relationships

        return Relationships(self)

    @relationships.setter
    def relationships(self, rels: typing.Mapping["Relationship", Iterable[_E]]):
        self._data().relationships = {
            relation.id: set(entity.id for entity in entities)
            for relation, entities in rels.items()
        }

    @property
    def replaced_by(self) -> _S:
        """`EntitySet`: A set of of replacements for an obsolete term.

        An obsolete entity can provide one or more replacement that can
        safely be used to automatically reassign instances to non-obsolete
        classes.

        See Also:
            `~Entity.consider`, which provides a set of entities suitable
            for replacement but requiring expert curation.

        """
        s = self._Set()
        s._ids = self._data().replaced_by
        s._ontology = self._ontology()
        return s

    @replaced_by.setter
    def replaced_by(self, replacements: Iterable[_E]) -> None:
        if isinstance(replacements, EntitySet):
            data = replacements._ids
        else:
            data = set(entity.id for entity in replacements)
        self._data().replaced_by = data

    @property
    def subsets(self) -> FrozenSet[str]:
        """`frozenset` of `str`: The subsets containing this entity."""
        return frozenset(self._data().subsets)

    @subsets.setter  # type: ignore
    @typechecked(property=True)
    def subsets(self, subsets: FrozenSet[str]):
        declared = set(s.name for s in self._ontology().metadata.subsetdefs)
        for subset in subsets:
            if subset not in declared:
                raise ValueError(f"undeclared subset: {subset!r}")
        self._data().subsets = set(subsets)

    @property
    def synonyms(self) -> FrozenSet[Synonym]:
        """`frozenset` of `Synonym`: A set of synonyms for this entity."""
        ontology, termdata = self._ontology(), self._data()
        return frozenset(Synonym(ontology, s) for s in termdata.synonyms)

    @synonyms.setter  # type: ignore
    @typechecked(property=True)
    def synonyms(self, synonyms: Iterable[Synonym]):
        self._data().synonyms = {syn._data() for syn in synonyms}

    @property
    def union_of(self) -> _S:
        s = self._Set()
        s._ids = self._data().union_of
        s._ontology = self._ontology()
        return s

    @union_of.setter
    def union_of(self, union_of: Iterable[_E]) -> None:
        if isinstance(union_of, EntitySet):
            data = union_of._ids
        else:
            data = set()
            for entity in union_of:
                if not isinstance(entity, Entity):
                    ty = type(entity).__name__
                    raise TypeError(f"expected `Entity`, found {ty}")
                data.add(entity.id)
        if len(data) == 1:
            raise ValueError("'union_of' cannot have a cardinality of 1")
        self._data().union_of = data

    @property
    def xrefs(self) -> FrozenSet[Xref]:
        """`frozenset` of `Xref`: A set of database cross-references.

        Xrefs can be used to describe an analogous entity in another
        vocabulary, such as a database or a semantic knowledge base.
        """
        return frozenset(self._data().xrefs)

    @xrefs.setter  # type: ignore
    @typechecked(property=True)
    def xrefs(self, xrefs: FrozenSet[Xref]):
        self._data().xrefs = set(xrefs)

    # --- Convenience methods ------------------------------------------------

    def add_synonym(
        self,
        description: str,
        scope: Optional[str] = None,
        type: Optional[SynonymType] = None,
        xrefs: Optional[Iterable[Xref]] = None,
    ) -> Synonym:
        """Add a new synonym to the current entity.

        Arguments:
            description (`str`): The alternate definition of the entity, or a
                related human-readable synonym.
            scope (`str` or `None`): An optional synonym scope. Must be either
                **EXACT**, **RELATED**, **BROAD** or **NARROW** if given.
            type (`~pronto.SynonymType` or `None`): An optional synonym type.
                Must be declared in the header of the current ontology.
            xrefs (iterable of `Xref`, or `None`): A collections of database
                cross-references backing the origin of the synonym.

        Raises:
            ValueError: when given an invalid synonym type or scope.

        Returns:
            `~pronto.Synonym`: A new synonym for the terms. The synonym is
            already added to the `Entity.synonyms` collection.

        """
        # check the type is declared in the current ontology
        if type is None:
            type_id: Optional[str] = None
        else:
            try:
                type_id = self._ontology().get_synonym_type(type.id).id
            except KeyError as ke:
                raise ValueError(f"undeclared synonym type {type.id!r}") from ke

        data = SynonymData(description, scope, type_id, xrefs=xrefs)
        self._data().synonyms.add(data)
        return Synonym(self._ontology(), data)


class EntitySet(typing.Generic[_E], typing.MutableSet[_E]):
    """A specialized mutable set to store `Entity` instances."""

    # --- Magic methods ------------------------------------------------------

    def __init__(self, entities: Optional[Iterable[_E]] = None) -> None:
        self._ids: Set[str] = set()
        self._ontology: "Optional[Ontology]" = None

        for entity in entities if entities is not None else ():
            if __debug__ and not isinstance(entity, Entity):
                err_msg = "'entities' must be iterable of Entity, not {}"
                raise TypeError(err_msg.format(type(entity).__name__))
            if self._ontology is None:
                self._ontology = entity._ontology()
            if self._ontology is not entity._ontology():
                raise ValueError("entities do not originate from the same ontology")
            self._ids.add(entity.id)

    def __contains__(self, other: object):
        if isinstance(other, Entity):
            return other.id in self._ids
        return False

    def __iter__(self) -> Iterator[_E]:
        return map(lambda t: self._ontology[t], iter(self._ids))  # type: ignore

    def __len__(self):
        return len(self._ids)

    def __repr__(self):
        ontology = self._ontology
        elements = (repr(ontology[id_]) for id_ in self._ids)
        return f"{type(self).__name__}({{{', '.join(elements)}}})"

    def __iand__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if isinstance(other, EntitySet):
            self._ids &= other._ids
        else:
            super().__iand__(other)
        if not self._ids:
            self._ontology = None
        return self

    def __and__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if isinstance(other, EntitySet):
            s = type(self)()
            s._ids = self._ids.__and__(other._ids)
            s._ontology = self._ontology if s._ids else None
        else:
            s = type(self)(super().__and__(other))
        return s

    def __ior__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if not isinstance(other, EntitySet):
            other = type(self)(other)
        self._ids |= other._ids
        self._ontology = self._ontology or other._ontology
        return self

    def __or__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if isinstance(other, EntitySet):
            s = type(self)()
            s._ids = self._ids.__or__(other._ids)
            s._ontology = self._ontology or other._ontology
        else:
            s = type(self)(super().__or__(other))
        return s

    def __isub__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if isinstance(other, EntitySet):
            self._ids -= other._ids
        else:
            super().__isub__(other)
        if not self._ids:
            self._ontology = None
        return self

    def __sub__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if isinstance(other, EntitySet):
            s = type(self)()
            s._ids = self._ids.__sub__(other._ids)
            s._ontology = self._ontology
        else:
            s = type(self)(super().__sub__(other))
        return s

    def __ixor__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if isinstance(other, EntitySet):
            self._ids ^= other._ids
            self._ontology = self._ontology or other._ontology
        else:
            super().__ixor__(other)
        if not self._ids:
            self._ontology = None
        return self

    def __xor__(self, other: AbstractSet[_E]) -> "EntitySet[_E]":
        if isinstance(other, EntitySet):
            s = type(self)()
            s._ids = self._ids.__xor__(other._ids)
            s._ontology = self._ontology or other._ontology
        else:
            s = type(self)(super().__xor__(other))
        if not s._ids:
            s._ontology = None
        return s

    # --- Methods ------------------------------------------------------------

    def add(self, entity: _E) -> None:
        if self._ontology is None:
            self._ontology = entity._ontology()
        elif self._ontology is not entity._ontology():
            raise ValueError("cannot use `Entity` instances from different `Ontology`")
        self._ids.add(entity.id)

    def clear(self) -> None:
        self._ids.clear()
        self._ontology = None

    def discard(self, entity: _E) -> None:
        self._ids.discard(entity.id)

    def pop(self) -> _E:
        id_ = self._ids.pop()
        entity = self._ontology[id_]  # type: ignore
        if not self._ids:
            self._ontology = None
        return entity  # type: ignore

    def remove(self, entity: _E):
        if self._ontology is not None and self._ontology is not entity._ontology():
            raise ValueError("cannot use `Entity` instances from different `Ontology`")
        self._ids.remove(entity.id)

    # --- Attributes ---------------------------------------------------------

    @property
    def ids(self) -> FrozenSet[str]:
        return frozenset(map(operator.attrgetter("id"), iter(self)))

    @property
    def alternate_ids(self) -> FrozenSet[str]:
        return frozenset(id for entity in self for id in entity.alternate_ids)

    @property
    def names(self) -> FrozenSet[str]:
        return frozenset(map(operator.attrgetter("name"), iter(self)))



# coding: utf-8

import functools
import typing
import weakref
from typing import Iterable, Optional, Set

from .utils.meta import roundrepr, typechecked
from .xref import Xref

if typing.TYPE_CHECKING:
    from .ontology import Ontology


_SCOPES = frozenset({"EXACT", "RELATED", "BROAD", "NARROW", None})
__all__ = ["SynonymType", "SynonymData", "Synonym"]


@roundrepr
@functools.total_ordering
class SynonymType(object):
    """A user-defined synonym type."""

    id: str
    description: str
    scope: Optional[str]

    __slots__ = ("__weakref__", "id", "description", "scope")

    @typechecked()
    def __init__(self, id: str, description: str, scope: Optional[str] = None):
        if scope not in _SCOPES:
            raise ValueError(f"invalid synonym scope: {scope}")
        self.id = id
        self.description = description
        self.scope = scope

    def __eq__(self, other):
        if isinstance(other, SynonymType):
            return self.id == other.id
        return False

    def __lt__(self, other):
        if isinstance(other, SynonymType):
            if self.id < other.id:
                return True
            return self.id == other.id and self.description < other.description
        return NotImplemented

    def __hash__(self):
        return hash((SynonymType, self.id))


@roundrepr
@functools.total_ordering
class SynonymData(object):
    """Internal data storage of `Synonym` information."""

    description: str
    scope: Optional[str]
    type: Optional[str]
    xrefs: Set[Xref]

    __slots__ = ("__weakref__", "description", "type", "xrefs", "scope")

    def __eq__(self, other):
        if isinstance(other, SynonymData):
            return self.description == other.description and self.scope == other.scope
        return False

    def __lt__(self, other):  # FIXME?
        if not isinstance(other, SynonymData):
            return NotImplemented
        if self.type is not None and other.type is not None:
            return (self.description, self.scope, self.type, frozenset(self.xrefs)) < (
                self.description,
                self.scope,
                other.type,
                frozenset(other.xrefs),
            )
        else:
            return (self.description, self.scope, frozenset(self.xrefs)) < (
                self.description,
                self.scope,
                frozenset(other.xrefs),
            )

    def __hash__(self):
        return hash((self.description, self.scope))

    def __init__(
        self,
        description: str,
        scope: Optional[str] = None,
        type: Optional[str] = None,
        xrefs: Optional[Iterable[Xref]] = None,
    ):
        if scope not in _SCOPES:
            raise ValueError(f"invalid synonym scope: {scope}")
        self.description = description
        self.scope = scope
        self.type = type
        self.xrefs = set(xrefs) if xrefs is not None else set()


@functools.total_ordering
class Synonym(object):
    """A synonym for an entity, with respect to the OBO terminology."""

    __ontology: "Ontology"

    if typing.TYPE_CHECKING:

        __data: "weakref.ReferenceType[SynonymData]"

        def __init__(self, ontology: "Ontology", data: "SynonymData"):
            self.__data = weakref.ref(data)
            self.__ontology = ontology

        def _data(self) -> SynonymData:
            rdata = self.__data()
            if rdata is None:
                raise RuntimeError("synonym data was deallocated")
            return rdata

    else:

        __slots__: Iterable[str] = ("__weakref__", "__ontology", "_data")

        def __init__(self, ontology: "Ontology", syndata: "SynonymData"):
            if syndata.type is not None:
                if not any(t.id == syndata.type for t in ontology.synonym_types()):
                    raise ValueError(f"undeclared synonym type: {syndata.type}")
            self._data = weakref.ref(syndata)
            self.__ontology = ontology

    def __eq__(self, other: object):
        if isinstance(other, Synonym):
            return self._data() == other._data()
        return False

    def __lt__(self, other: object):
        if not isinstance(other, Synonym):
            return False
        return self._data().__lt__(other._data())

    def __hash__(self):
        return hash(self._data())

    def __repr__(self):
        return roundrepr.make(
            "Synonym",
            self.description,
            scope=(self.scope, None),
            type=(self.type, None),
            xrefs=(self.xrefs, set()),
        )

    @property
    def description(self) -> str:
        return self._data().description

    @description.setter  # type: ignore
    @typechecked(property=True)
    def description(self, description: str) -> None:
        self._data().description = description

    @property
    def type(self) -> Optional[SynonymType]:
        ontology, syndata = self.__ontology, self._data()
        if syndata.type is not None:
            return next(t for t in ontology.synonym_types() if t.id == syndata.type)
        return None

    @type.setter  # type: ignore
    @typechecked(property=True)
    def type(self, type_: Optional[SynonymType]) -> None:
        synonyms: Iterable[SynonymType] = self.__ontology.synonym_types()
        if type_ is not None and not any(type_.id == s.id for s in synonyms):
            raise ValueError(f"undeclared synonym type: {type_.id}")
        self._data().type = type_.id if type_ is not None else None

    @property
    def scope(self) -> Optional[str]:
        return self._data().scope

    @scope.setter  # type: ignore
    @typechecked(property=True)
    def scope(self, scope: Optional[str]):
        if scope not in _SCOPES:
            raise ValueError(f"invalid synonym scope: {scope}")
        self._data().scope = scope

    @property
    def xrefs(self) -> Set[Xref]:
        return self._data().xrefs

    @xrefs.setter
    def xrefs(self, xrefs: Iterable[Xref]):
        self._data().xrefs = set(xrefs)



import io

from ..relationship import Relationship
from ..term import Term
from ._fastobo import FastoboSerializer
from .base import BaseSerializer


class OboSerializer(FastoboSerializer, BaseSerializer):

    format = "obo"

    def dump(self, file):
        writer = io.TextIOWrapper(file)
        try:
            # dump the header
            if self.ont.metadata:
                header = self._to_header_frame(self.ont.metadata)
                file.write(str(header).encode("utf-8"))
                if self.ont._terms.entities or self.ont._relationships.entities:
                    file.write(b"\n")
            # dump terms
            if self.ont._terms.entities:
                for i, id in enumerate(sorted(self.ont._terms.entities)):
                    data = self.ont._terms.entities[id]
                    frame = self._to_term_frame(Term(self.ont, data))
                    file.write(str(frame).encode("utf-8"))
                    if (
                        i < len(self.ont._terms.entities) - 1
                        or self.ont._relationships.entities
                    ):
                        file.write(b"\n")
            # dump typedefs
            if self.ont._relationships.entities:
                for i, id in enumerate(sorted(self.ont._relationships.entities)):
                    data = self.ont._relationships.entities[id]
                    frame = self._to_typedef_frame(Relationship(self.ont, data))
                    file.write(str(frame).encode("utf-8"))
                    if i < len(self.ont._relationships.entities) - 1:
                        file.write(b"\n")
        finally:
            writer.detach()



from collections.abc import Sized
from typing import Generic, Iterator, TypeVar

__all__ = ["SizedIterator"]
S = TypeVar("S")
T = TypeVar("T")


class SizedIterator(Generic[T], Iterator[T], Sized):
    """A wrapper for iterators which length is known in advance."""

    def __init__(self, it: Iterator[T], length: int):
        self._it = it
        self._length = length

    def __len__(self) -> int:
        return self._length

    def __length_hint__(self) -> int:
        return self._length

    def __iter__(self: S) -> S:
        return self

    def __next__(self) -> T:
        val = next(self._it)
        self._length -= 1
        return val



import multiprocessing.pool
import os

import fastobo

from ..utils.meta import typechecked
from ._fastobo import FastoboParser
from .base import BaseParser


class OboParser(FastoboParser, BaseParser):
    @classmethod
    def can_parse(cls, path, buffer):
        return buffer.lstrip().startswith((b"format-version:", b"[Term", b"[Typedef"))

    def parse_from(self, handle, threads=None):
        # Load the OBO document through an iterator using fastobo
        doc = fastobo.iter(handle, ordered=True)

        # Extract metadata from the OBO header
        with typechecked.disabled():
            self.ont.metadata = self.extract_metadata(doc.header())

        # Resolve imported dependencies
        self.ont.imports.update(
            self.process_imports(
                self.ont.metadata.imports,
                self.ont.import_depth,
                os.path.dirname(self.ont.path or str()),
                self.ont.timeout,
                threads=threads,
            )
        )

        # Merge lineage cache from imports
        self.import_lineage()

        # Extract frames from the current document.
        with typechecked.disabled():
            try:
                with multiprocessing.pool.ThreadPool(threads) as pool:
                    pool.map(self.extract_entity, doc)
            except SyntaxError as s:
                location = self.ont.path, s.lineno, s.offset, s.text
                raise SyntaxError(s.args[0], location) from None

        # Update lineage cache with symmetric of `subClassOf`
        self.symmetrize_lineage()



import datetime
import functools
import typing
from typing import Dict, List, Optional, Set, Tuple

from .pv import PropertyValue
from .synonym import SynonymType
from .utils.meta import roundrepr, typechecked

__all__ = ["Subset", "Metadata"]


@roundrepr
@functools.total_ordering
class Subset(object):
    """A definition of a subset in an ontology.

    Attributes:
        name (`str`): The name of the subset, as an OBO short identifier.
        description (`str`): A description of the subset, as defined in the
            metadata part of the ontology file.

    """

    name: str
    description: str

    __slots__ = ("__weakref__", "name", "description")

    if typing.TYPE_CHECKING:
        __annotations__: Dict[str, str]

    @typechecked()
    def __init__(self, name: str, description: str):
        """Create a new subset with the given name and description."""
        self.name: str = name
        self.description: str = description

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Subset):
            return self.name == other.name
        return False

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Subset):
            return typing.cast(bool, NotImplemented)
        return self.name < other.name

    def __hash__(self) -> int:
        return hash((Subset, self.name))


class Metadata(object):
    """A mapping containing metadata about the current ontology.

    Attributes:
        format_version (`str`): The OBO format version of the referenced
            ontology. **1.4** is the default since ``pronto`` can only
            parse and write OBO documents of that format version.
        data_version (`str` or `None`): The OBO data version of the ontology,
            which is then expanded to the ``versionIRI`` if translated to
            OWL.
        ontology (`str` or `None`): The identifier of the ontology, either as
            a short OBO identifier or as a full IRI.
        date (`~datetime.datetime` or `None`): The date the ontology was last
            modified, if any.
        default_namespace (`str` or `None`): The default namespace to use for
            entity frames lacking a ``namespace`` clause.
        namespace_id_rule (`str` or `None`): The production rule for
            identifiers in the current ontology document. *soft-deprecated,
            used mostly by OBO-Edit or other outdated tools*.
        owl_axioms (`list` of `str`): A list of OWL axioms that cannot be
            expressed in OBO language, serialized in OWL2 Functional syntax.
        saved_by (`str` or `None`): The name of the person that last saved the
            ontology file.
        auto_generated_by (`str` or `None`): The name of the software that was
            used to generate the file.
        subsetdefs (`set` of `str`): A set of ontology subsets declared in the
            ontology files.
        imports (`set` of `str`): A set of references to other ontologies that
            are imported by the current ontology. OBO requires all entities
            referenced in the file to be reachable through imports (excluding
            databases cross-references).
        synonymtypedefs (`set` of `~pronto.SynonymType`): A set of user-defined
            synonym types including a description and an optional scope.
        idspaces (`dict` of `str` to couple of `str`): A mapping between a
            local ID space and a global ID space, with an optional description
            of the mapping.
        remarks (`set` of `str`): A set of general comments for this file,
            which will be preserved by a parser/serializer as opposed to
            inline comments using ``!``.
        annotations (`set` of `PropertyValue`): A set of annotations relevant
            to the whole file. OBO property-values are semantically equivalent
            to  ``owl:AnnotationProperty`` in OWL2.
        unreserved (`dict` of `str` to `set` of `str`): A mapping of
            unreserved tags to values found in the ontology header.

    """

    format_version: Optional[str]
    data_version: Optional[str]
    ontology: Optional[str]
    date: Optional[datetime.datetime]
    default_namespace: Optional[str]
    namespace_id_rule: Optional[str]
    owl_axioms: List[str]
    saved_by: Optional[str]
    auto_generated_by: Optional[str]
    subsetdefs: Set[Subset]
    imports: Set[str]
    synonymtypedefs: Set[SynonymType]
    idspaces: Dict[str, Tuple[str, Optional[str]]]  # FIXME: better typing?
    remarks: Set[str]
    annotations: Set[PropertyValue]
    unreserved: Dict[str, Set[str]]

    if typing.TYPE_CHECKING:
        __annotations__: Dict[str, str]

    def __init__(
        self,
        format_version: Optional[str] = "1.4",
        data_version: Optional[str] = None,
        ontology: Optional[str] = None,
        date: Optional[datetime.datetime] = None,
        default_namespace: Optional[str] = None,
        namespace_id_rule: Optional[str] = None,
        owl_axioms: Optional[List[str]] = None,
        saved_by: Optional[str] = None,
        auto_generated_by: Optional[str] = None,
        subsetdefs: Set[Subset] = None,
        imports: Optional[Dict[str, str]] = None,
        synonymtypedefs: Set[SynonymType] = None,
        idspaces: Dict[str, Tuple[str, Optional[str]]] = None,
        remarks: Set[str] = None,
        annotations: Set[PropertyValue] = None,
        **unreserved: Set[str],
    ):
        self.format_version = format_version
        self.data_version = data_version
        self.ontology = ontology
        self.date = date
        self.default_namespace = default_namespace
        self.namespace_id_rule = namespace_id_rule
        self.owl_axioms = owl_axioms or list()
        self.saved_by = saved_by
        self.auto_generated_by = auto_generated_by
        self.subsetdefs = set(subsetdefs) if subsetdefs is not None else set()
        self.imports = set(imports) if imports is not None else set()
        self.synonymtypedefs = (
            set(synonymtypedefs) if synonymtypedefs is not None else set()
        )
        self.idspaces = idspaces or dict()
        self.remarks = remarks or set()
        self.annotations = annotations or set()
        self.unreserved = unreserved

    def __bool__(self) -> bool:
        """Return `False` if the instance does not contain any metadata."""
        return any(bool(getattr(self, x)) for x in self.__annotations__)



import typing
from typing import Iterable, Iterator

from ..utils.meta import typechecked
from . import Entity, EntitySet

if typing.TYPE_CHECKING:
    from ..relationship import Relationship

_E = typing.TypeVar("_E", bound=Entity)
_S = typing.TypeVar("_S", bound=EntitySet)


class Relationships(typing.MutableMapping["Relationship", _S], typing.Generic[_E, _S]):
    """A dedicated mutable mapping to manage the relationships of an entity."""

    def __init__(self, entity: _E):
        self._inner = entity._data().relationships
        self._entity = entity
        self._ontology = entity._ontology()

    def __getitem__(self, item: "Relationship") -> _S:
        if item.id not in self._inner:
            raise KeyError(item)
        s = self._entity._Set()
        s._ids = self._inner[item.id]
        s._ontology = self._ontology
        return s

    def __delitem__(self, item: "Relationship"):
        if item.id not in self._inner:
            raise KeyError(item)
        del self._inner[item.id]

    def __len__(self) -> int:
        return len(self._inner)

    def __iter__(self) -> Iterator["Relationship"]:
        from ..relationship import Relationship

        return (self._ontology.get_relationship(id_) for id_ in self._inner)

    def __setitem__(self, key: "Relationship", entities: Iterable[_E]):
        if key._ontology() is not self._ontology:
            raise ValueError("cannot use a relationship from a different ontology")
        self._inner[key.id] = {entity.id for entity in entities}



import collections.abc
import datetime
import itertools
import operator
import typing
import warnings
from typing import (
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from . import relationship
from .definition import Definition
from .entity import Entity, EntityData, EntitySet
from .logic import SubclassesIterator, SuperclassesIterator
from .logic.lineage import SubclassesHandler, SuperclassesHandler
from .pv import PropertyValue
from .relationship import Relationship
from .synonym import SynonymData
from .utils.meta import typechecked
from .utils.warnings import NotImplementedWarning
from .xref import Xref

if typing.TYPE_CHECKING:
    from .ontology import Ontology


__all__ = ["TermData", "Term", "TermSet"]


class TermData(EntityData):  # noqa: R0902, R0903
    """Internal data storage of `Term` information."""

    id: str
    anonymous: bool
    name: Optional[str]
    alternate_ids: Set[str]
    definition: Optional[Definition]
    comment: Optional[str]
    synonyms: Set[SynonymData]
    subsets: Set[str]
    namespace: Optional[str]
    xrefs: Set[Xref]
    intersection_of: Set[Union[str, Tuple[str, str]]]
    obsolete: bool
    builtin: bool
    created_by: Optional[str]
    creation_date: Optional[datetime.datetime]
    annotations: Set[PropertyValue]

    if typing.TYPE_CHECKING:
        __annotations__: Dict[str, str]

    __slots__ = tuple(__annotations__)  # noqa: E0602

    def __init__(
        self,
        id,
        anonymous=False,
        name=None,
        alternate_ids=None,
        definition=None,
        comment=None,
        synonyms=None,
        subsets=None,
        namespace=None,
        xrefs=None,
        intersection_of=None,
        union_of=None,
        disjoint_from=None,
        relationships=None,
        obsolete=False,
        replaced_by=None,
        consider=None,
        builtin=False,
        created_by=None,
        creation_date=None,
        equivalent_to=None,
        annotations=None,
    ):
        self.id = id
        self.anonymous = anonymous
        self.name = name
        self.alternate_ids = alternate_ids or set()
        self.definition = definition
        self.comment = comment
        self.synonyms = synonyms or set()
        self.subsets = subsets or set()
        self.namespace = namespace or None
        self.xrefs = xrefs or set()
        self.intersection_of = intersection_of or set()
        self.union_of = union_of or set()
        self.disjoint_from = disjoint_from or set()
        self.relationships = relationships or dict()
        self.obsolete = obsolete
        self.replaced_by = replaced_by or set()
        self.consider = consider or set()
        self.builtin = builtin
        self.created_by = created_by
        self.creation_date = creation_date
        self.equivalent_to = equivalent_to or set()
        self.annotations = annotations or set()


class TermSet(EntitySet["Term"]):
    """A specialized mutable set to store `Term` instances."""

    # --- Methods ------------------------------------------------------------

    def subclasses(
        self, distance: Optional[int] = None, with_self: bool = True
    ) -> SubclassesIterator:
        """Get an iterator over the subclasses of all terms in the set.

        Caution:
            Contrary to `pronto.Term.subclasses`, this method **does not**
            return a handler that lets you edit the subclasses directly.
            Adding a new subclass to all the members of the set must be
            done explicitly.

        """
        return SubclassesIterator(*self, distance=distance, with_self=with_self)

    def superclasses(
        self, distance: Optional[int] = None, with_self: bool = True
    ) -> SuperclassesIterator:
        """Get an iterator over the superclasses of all terms in the set.

        Caution:
            Contrary to `pronto.Term.superclasses`, this method **does not**
            return a handler that lets you edit the superclasses directly.
            Adding a new superclass to all the members of the set must be
            done explicitly.

        Example:
            >>> ms = pronto.Ontology("ms.obo")
            >>> s = pronto.TermSet({ms['MS:1000122'], ms['MS:1000124']})
            >>> s.superclasses(with_self=False).to_set().ids
            frozenset({'MS:1000031'})
            >>> ms["MS:1000031"]
            Term('MS:1000031', name='instrument model')

        """
        return SuperclassesIterator(*self, distance=distance, with_self=with_self)


class Term(Entity["TermData", "TermSet"]):
    """A term, corresponding to a node in the ontology graph.

    Formally a `Term` frame is equivalent to an ``owl:Class`` declaration in
    OWL2 language. However, some constructs may not be possible to express in
    both OBO and OWL2.

    `Term` should not be manually instantiated, but obtained from an existing
    `Ontology` instance, using either the `~Ontology.create_term` or the
    `~Ontology.get_term` method.
    """

    if typing.TYPE_CHECKING:

        def __init__(self, ontology: "Ontology", termdata: "TermData"):
            super().__init__(ontology, termdata)

        def _data(self) -> "TermData":
            return typing.cast("TermData", super()._data())

    # --- Associated type variables ------------------------------------------

    _Set = TermSet
    _data_getter = operator.attrgetter("_terms")

    # --- Methods ------------------------------------------------------------

    def objects(self, r: Relationship) -> Iterator["Term"]:
        """Iterate over the terms ``t`` verifying ``self · r · t``.

        Example:
            >>> go = pronto.Ontology.from_obo_library("go.obo")
            >>> go['GO:0048870']
            Term('GO:0048870', name='cell motility')
            >>> list(go['GO:0048870'].objects(go.get_relationship('part_of')))
            [Term('GO:0051674', name='localization of cell')]

        Todo:
            Make `Term.objects` take in account ``holds_over_chain`` and
            ``transitive_over`` values of the relationship it is building an
            iterator with.

        """
        # delegate import of NetworkX
        import networkx

        if r._data() is relationship._BUILTINS["is_a"]:
            warnings.warn(
                "using the `is_a` relationship with `Term.objects` will not be "
                "supported in future versions, use `Term.superclasses` instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return self.superclasses()

        warnings.warn(
            "`Term.objects` is not semantically correct, most of the logic "
            "rules have not been implemented. Consider using an actual "
            "reasoner instead.",
            category=NotImplementedWarning,
            stacklevel=2,
        )

        g = networkx.MultiDiGraph()
        ont = self._ontology()

        # Build the graph
        for t in ont.terms():
            g.add_node(t.id)
            for (rel, terms) in t.relationships.items():
                for t2 in terms:
                    g.add_edge(t.id, t2.id, key=rel.id)
                    if rel.symmetric:
                        g.add_edge(t2.id, t.id, key=rel.id)
                    elif rel.inverse_of is not None:
                        g.add_edge(t2.id, t.id, key=rel.inverse_of.id)

        # Search objects terms
        red, done = set(), set()
        is_red = red.__contains__
        frontier = {self.id}

        # Initial connected components
        if r.reflexive:
            red.add(self.id)
            yield self
        for other in g.neighbors(self.id):
            if r.id in g.get_edge_data(self.id, other):
                red.add(other)
                yield ont.get_term(other)

        # Explore the graph
        while frontier:
            node = frontier.pop()
            frontier.update(n for n in g.neighbors(node) if n not in done)
            if is_red(node) and r.transitive:
                for other in itertools.filterfalse(is_red, g.neighbors(node)):
                    if r.id in g.get_edge_data(node, other):
                        red.add(other)
                        yield ont.get_term(other)
            done.add(node)

    @typechecked()
    def superclasses(
        self,
        distance: Optional[int] = None,
        with_self: bool = True,
    ) -> "SuperclassesHandler":
        """Get an handle over the superclasses of this `Term`.

        In order to follow the semantics of ``rdf:subClassOf``, which in turn
        respects the mathematical definition of subset inclusion, ``is_a`` is
        defined as a reflexive relationship, and so is its inverse
        relationship.

        Arguments:
            distance (int, optional): The maximum distance between this term
                and the yielded superclass (`0` for the term itself, `1` for
                its immediate superclasses, etc.). Use `None` to explore
                transitively the entire directed graph.
            with_self (bool): Whether or not to include the current term in
                the terms being yielded. RDF semantics state that the
                ``rdfs:subClassOf`` property is reflexive, so this is enabled
                by default, but in most practical cases only the distinct
                subclasses are desired.

        Yields:
            `Term`: Superclasses of the selected term, breadth-first. The
            first element is always the term itself, use `itertools.islice`
            to skip it.

        Example:
            >>> ms = pronto.Ontology.from_obo_library("ms.obo")
            >>> sup = iter(ms['MS:1000143'].superclasses())
            >>> next(sup)
            Term('MS:1000143', name='API 150EX')
            >>> next(sup)
            Term('MS:1000121', name='SCIEX instrument model')
            >>> next(sup)
            Term('MS:1000031', name='instrument model')

        Note:
            The time complexity for this algorithm is in :math:`O(n)`, where
            :math:`n` is the number of subclasses of initial term.

        See Also:
            The `RDF Schema 1.1 <https://www.w3.org/TR/rdf-schema/>`_
            specification, defining the ``rdfs:subClassOf`` property, which
            the ``is_a`` relationship is translated to in OWL2 language.

        """
        return SuperclassesHandler(self, distance=distance, with_self=with_self)

    @typechecked()
    def subclasses(
        self,
        distance: Optional[int] = None,
        with_self: bool = True,
    ) -> "SubclassesHandler":
        """Get an handle over the subclasses of this `Term`.

        Arguments:
            distance (int, optional): The maximum distance between this term
                and the yielded subclass (`0` for the term itself, `1` for
                its immediate children, etc.). Use `None` to explore the
                entire directed graph transitively.
            with_self (bool): Whether or not to include the current term in
                the terms being yielded. RDF semantics state that the
                ``rdfs:subClassOf`` property is reflexive, so this is enabled
                by default, but in most practical cases only the distinct
                subclasses are desired.

        Yields:
            `Term`: Subclasses of the selected term, breadth-first. The first
            element is always the term itself, use `itertools.islice` to skip
            it.

        Example:
            >>> ms = pronto.Ontology.from_obo_library("ms.obo")
            >>> sub = iter(ms['MS:1000031'].subclasses())
            >>> next(sub)
            Term('MS:1000031', name='instrument model')
            >>> next(sub)
            Term('MS:1000121', name='SCIEX instrument model')
            >>> next(sub)
            Term('MS:1000122', name='Bruker Daltonics instrument model')

        Hint:
            Use the ``to_set`` method of the returned iterator to efficiently
            collect all subclasses into a `TermSet`.

        Note:
            This method has a runtime that is :math:`O(n)` where :math:`n` is
            the number of subclasses of the initial term. While OBO and OWL
            only explicit the *superclassing* relationship (equivalent to the
            ``rdfs:subClassOf`` property in RDF), we can build a cache that
            stores the edges of the resulting knowledge graph in an index
            accessible by both endpoints of each edge.

        """
        return SubclassesHandler(self, distance=distance, with_self=with_self)

    def is_leaf(self) -> bool:
        """Check whether the term is a leaf in the ontology.

        We define leaves as nodes in the ontology which do not have subclasses
        since the subclassing relationship is directed and can be used to
        create a DAG of all the terms in the ontology.

        Example:
            >>> ms = pronto.Ontology.from_obo_library("ms.obo")
            >>> ms['MS:1000031'].is_leaf()   # instrument model
            False
            >>> ms['MS:1001792'].is_leaf()   # Xevo TQ-S
            True

        Note:
            This method has a runtime of :math:`O(1)` as `Ontology` objects
            internally cache the subclasses of each term.

        """
        return not self._ontology()._terms.lineage[self.id].sub

    # --- Attributes ---------------------------------------------------------

    @property
    def intersection_of(self) -> FrozenSet[Union["Term", Tuple[Relationship, "Term"]]]:
        """`frozenset`: The terms this term is an intersection of."""
        ont, termdata = self._ontology(), self._data()
        intersection_of: List[Union["Term", Tuple[Relationship, "Term"]]] = []
        for item in termdata.intersection_of:
            if isinstance(item, str):
                intersection_of.append(ont.get_term(typing.cast(str, item)))
            else:
                r, t = item
                intersection_of.append((ont.get_relationship(r), ont.get_term(t)))
        return frozenset(intersection_of)

    @intersection_of.setter
    def intersection_of(
        self, intersection_of: Iterable[Union["Term", Tuple[Relationship, "Term"]]]
    ):
        data = set()
        for item in intersection_of:
            if isinstance(item, Term):
                data.add(item.id)
            elif isinstance(item, collections.abc.Collection) and len(item) == 2:
                rel, term = item  # type: ignore
                data.add((rel.id, term.id))  # type: ignore
            else:
                msg = "expected iterable of `Term` or `Relationship`, `Term` couple, found: {}"
                raise TypeError(msg.format(type(item).__name__))
        self._data().intersection_of = data



"""Object hierarchy of property-value annotations in OBO files.
"""

import abc
import functools

import fastobo

from .utils.meta import roundrepr, typechecked


__all__ = ["PropertyValue", "LiteralPropertyValue", "ResourcePropertyValue"]


class PropertyValue(abc.ABC):
    """A property-value, which adds annotations to an entity."""

    property: str
    __slots__ = ("property",)

    @abc.abstractmethod
    def __eq__(self, other: object):
        return NotImplemented

    @abc.abstractmethod
    def __lt__(self, other: object):
        return NotImplemented


@roundrepr
@functools.total_ordering
class LiteralPropertyValue(PropertyValue):
    """A property-value which adds a literal annotation to an entity."""

    literal: str
    datatype: str
    __slots__ = ("literal", "datatype")

    @typechecked()
    def __init__(self, property: str, literal: str, datatype: str = "xsd:string"):
        """Create a new `LiteralPropertyValue` instance.

        Arguments:
            property (str): The annotation property, as an OBO identifier.
            literal (str): The serialized value of the annotation.
            datatype (str): The datatype of the annotation property value.
                Defaults to `xsd:string`.

        """
        if not fastobo.id.is_valid(property):
            raise ValueError("invalid identifier: {}".format(property))
        if not fastobo.id.is_valid(datatype):
            raise ValueError("invalid identifier: {}".format(datatype))

        self.property = property
        self.literal = literal
        self.datatype = datatype

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LiteralPropertyValue):
            return (
                self.property == other.property
                and self.literal == other.literal
                and self.datatype == other.datatype
            )
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, LiteralPropertyValue):
            return (self.property, self.literal, self.datatype) < (
                other.property,
                other.literal,
                other.datatype,
            )
        elif isinstance(other, ResourcePropertyValue):
            return self.property < other.property
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash((LiteralPropertyValue, self.property, self.literal, self.datatype))


@roundrepr
@functools.total_ordering
class ResourcePropertyValue(PropertyValue):
    """A property-value which adds a resource annotation to an entity."""

    resource: str
    __slots__ = ("resource",)

    @typechecked()
    def __init__(self, property: str, resource: str):
        """Create a new `ResourcePropertyValue` instance.

        Arguments:
            property (str): The annotation property, as an OBO identifier.
            resource (str): The annotation entity value, as an OBO identifier.

        """
        if not fastobo.id.is_valid(property):
            raise ValueError("invalid identifier: {}".format(property))
        if not fastobo.id.is_valid(resource):
            raise ValueError("invalid identifier: {}".format(resource))

        self.property = property
        self.resource = resource

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ResourcePropertyValue):
            return (self.property, self.resource) == (other.property, other.resource)
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, ResourcePropertyValue):
            return (self.property, self.resource) < (other.property, other.resource)
        elif isinstance(other, LiteralPropertyValue):
            return self.property < other.property
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash((LiteralPropertyValue, self.property, self.resource))
